# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.utils import is_opencv_available
from unitorch.cli import hf_endpoint_url

__hf_hub_stable_v1_5_dict__ = lambda name: {
    "unet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/unet/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/unet/diffusion_pytorch_model.bin"
        ),
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/pytorch_model.bin"
        ),
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_v1_5_safetensors_dict__ = lambda name: {
    "unet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/unet/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/unet/diffusion_pytorch_model.safetensors"
        ),
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/model.safetensors"
        ),
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_v2_dict__ = __hf_hub_stable_v1_5_dict__
__hf_hub_stable_v2_safetensors_dict__ = __hf_hub_stable_v1_5_safetensors_dict__
__hf_hub_stable_v2_1_dict__ = __hf_hub_stable_v1_5_dict__
__hf_hub_stable_v2_1_safetensors_dict__ = __hf_hub_stable_v1_5_safetensors_dict__

__hf_hub_stable_xl_dict__ = lambda name: {
    "unet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/unet/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/unet/diffusion_pytorch_model.bin"
        ),
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/pytorch_model.bin"
        ),
    },
    "text2": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_2/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder_2/pytorch_model.bin"
        ),
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_xl_safetensors_dict__ = lambda name: {
    "unet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/unet/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/unet/diffusion_pytorch_model.fp16.safetensors"
        ),
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/model.fp16.safetensors"
        ),
    },
    "text2": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_2/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder_2/model.fp16.safetensors"
        ),
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_3_safetensors_dict__ = lambda name: {
    "transformer": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/transformer/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/transformer/diffusion_pytorch_model.safetensors"
        ),
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/model.safetensors"
        ),
    },
    "text2": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_2/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder_2/model.safetensors"
        ),
    },
    "text3": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_3/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_3/spiece.model"),
        "weight": [
            hf_endpoint_url(
                f"/{name}/resolve/main/text_encoder_3/model-{str(i).rjust(5, '0')}-of-00002.safetensors"
            )
            for i in range(1, 3)
        ],
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_flux_safetensors_dict__ = lambda name, n1=3, n2=2: {
    "transformer": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/transformer/config.json"),
        "weight": [
            hf_endpoint_url(
                f"{name}/resolve/main/transformer/diffusion_pytorch_model-{str(i).rjust(5, '0')}-of-{str(n1).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n1 + 1)
        ],
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/model.safetensors"
        ),
    },
    "text2": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_2/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/spiece.model"),
        "weight": [
            hf_endpoint_url(
                f"{name}/resolve/main/text_encoder_2/model-{str(i).rjust(5, '0')}-of-{str(n2).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n2 + 1)
        ],
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_stable_flux_ctrl_safetensors_dict__ = lambda name, n1=3, n2=4: {
    "transformer": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/transformer/config.json"),
        "weight": [
            hf_endpoint_url(
                f"{name}/resolve/main/transformer/diffusion_pytorch_model-{str(i).rjust(5, '0')}-of-{str(n1).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n1 + 1)
        ],
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/vocab.json"),
        "merge": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/merges.txt"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/text_encoder/model.safetensors"
        ),
    },
    "text2": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder_2/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer_2/spiece.model"),
        "weight": [
            hf_endpoint_url(
                f"{name}/resolve/main/text_encoder_2/model-{str(i).rjust(5, '0')}-of-{str(n2).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n2 + 1)
        ],
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_wan_v2_1_safetensors_dict__ = lambda name, n1=2, n2=5, im=False: {
    "transformer": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/transformer/config.json"),
        "weight": [
            hf_endpoint_url(
                f"/{name}/resolve/main/transformer/diffusion_pytorch_model-{str(i).rjust(5, '0')}-of-{str(n1).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n1 + 1)
        ],
    },
    "text": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/text_encoder/config.json"),
        "vocab": hf_endpoint_url(f"/{name}/resolve/main/tokenizer/spiece.model"),
        "weight": [
            hf_endpoint_url(
                f"/{name}/resolve/main/text_encoder/model-{str(i).rjust(5, '0')}-of-{str(n2).rjust(5, '0')}.safetensors"
            )
            for i in range(1, n2 + 1)
        ],
    },
    "image": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/image_encoder/config.json"),
        "vision_config": hf_endpoint_url(
            f"/{name}/resolve/main/image_processor/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/image_encoder/model.safetensors"
        ),
    }
    if im
    else {},
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_hub_vae_dict = lambda name: {
    "vae": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/vae/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/vae/diffusion_pytorch_model.bin"
        ),
    },
}

__hf_hub_vae_safetensors_dict__ = lambda name: {
    "vae": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/vae/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/vae/diffusion_pytorch_model.safetensors"
        ),
    },
}

__hf_hub_vae_safetensors_fp16_dict__ = lambda name: {
    "vae": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/vae/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/vae/diffusion_pytorch_model.fp16.safetensors"
        ),
    },
}

__hf_hub_controlnet_dict__ = lambda name: {
    "controlnet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/config.json"),
        "weight": hf_endpoint_url(f"/{name}/resolve/main/diffusion_pytorch_model.bin"),
    }
}

__hf_hub_controlnet_safetensors_dict__ = lambda name: {
    "controlnet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/diffusion_pytorch_model.safetensors"
        ),
    }
}

__hf_hub_adapter_dict__ = lambda name: {
    "adapter": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/config.json"),
        "weight": hf_endpoint_url(f"/{name}/resolve/main/diffusion_pytorch_model.bin"),
    }
}

__hf_hub_adapter_safetensors_dict__ = lambda name: {
    "adapter": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/diffusion_pytorch_model.safetensors"
        ),
    }
}

__hf_hub_stable_video_safetensors_dict__ = lambda name: {
    "unet": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/unet/config.json"),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/unet/diffusion_pytorch_model.safetensors"
        ),
    },
    "image": {
        "config": hf_endpoint_url(f"/{name}/resolve/main/image_encoder/config.json"),
        "vision_config": hf_endpoint_url(
            f"/{name}/resolve/main/feature_extractor/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            f"/{name}/resolve/main/image_encoder/model.safetensors"
        ),
    },
    "scheduler": hf_endpoint_url(
        f"/{name}/resolve/main/scheduler/scheduler_config.json"
    ),
}

__hf_flux_dev_redux_safetensors_dict__ = lambda x: {
    "image": {
        "config": hf_endpoint_url(f"/{x}/resolve/main/image_encoder/config.json"),
        "vision_config": hf_endpoint_url(
            f"/{x}/resolve/main/feature_extractor/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(f"/{x}/resolve/main/image_encoder/model.safetensors"),
    },
    "redux_image": {
        "config": hf_endpoint_url(f"/{x}/resolve/main/image_embedder/config.json"),
        "weight": hf_endpoint_url(
            f"/{x}/resolve/main/image_embedder/diffusion_pytorch_model.safetensors"
        ),
    },
}

pretrained_stable_infos = {
    "stable-v1.5": {
        **__hf_hub_stable_v1_5_dict__("botp/stable-diffusion-v1-5"),
        **__hf_hub_vae_dict("botp/stable-diffusion-v1-5"),
    },
    "stable-v1.5-realistic-v5.1-no-vae": {
        **__hf_hub_stable_v1_5_safetensors_dict__(
            "SG161222/Realistic_Vision_V5.1_noVAE"
        ),
        **__hf_hub_vae_safetensors_dict__("SG161222/Realistic_Vision_V5.1_noVAE"),
    },
    "stable-v1.5-realistic-v5.1": {
        **__hf_hub_stable_v1_5_dict__("stablediffusionapi/realistic-vision-v51"),
        **__hf_hub_vae_dict("stablediffusionapi/realistic-vision-v51"),
    },
    "stable-v1.5-film": {
        **__hf_hub_stable_v1_5_safetensors_dict__("Yntec/Film"),
        **__hf_hub_vae_safetensors_dict__("Yntec/Film"),
    },
    "stable-v1.5-cyber-realistic": {
        **__hf_hub_stable_v1_5_safetensors_dict__("Yntec/CyberRealistic"),
        **__hf_hub_vae_safetensors_dict__("Yntec/CyberRealistic"),
    },
    "stable-v1.5-majicmix-realistic-v6": {
        **__hf_hub_stable_v1_5_dict__("digiplay/majicMIX_realistic_v6"),
        **__hf_hub_vae_dict("digiplay/majicMIX_realistic_v6"),
    },
    "stable-v1.5-nitrosocke-ghibli": {
        **__hf_hub_stable_v1_5_dict__("nitrosocke/Ghibli-Diffusion"),
        **__hf_hub_vae_dict("nitrosocke/Ghibli-Diffusion"),
    },
    "stable-v1.5-inpainting": {
        **__hf_hub_stable_v1_5_dict__("botp/stable-diffusion-v1-5-inpainting"),
        **__hf_hub_vae_dict("botp/stable-diffusion-v1-5-inpainting"),
    },
    "stable-v1.5-realistic-v5.1-inpainting": {
        **__hf_hub_stable_v1_5_safetensors_dict__(
            "Uminosachi/realisticVisionV51_v51VAE-inpainting"
        ),
        **__hf_hub_vae_safetensors_dict__(
            "Uminosachi/realisticVisionV51_v51VAE-inpainting"
        ),
    },
    "stable-v1.5-dreamshaper-8-inpainting": {
        **__hf_hub_stable_v1_5_safetensors_dict__("Lykon/dreamshaper-8-inpainting"),
        **__hf_hub_vae_safetensors_dict__("Lykon/dreamshaper-8-inpainting"),
    },
    "stable-v1.5-anything-4.0-inpainting": {
        **__hf_hub_stable_v1_5_safetensors_dict__("Sanster/anything-4.0-inpainting"),
        **__hf_hub_vae_safetensors_dict__("Sanster/anything-4.0-inpainting"),
    },
    "stable-v1.5-x4-upscaler": {
        **__hf_hub_stable_v1_5_dict__("stabilityai/stable-diffusion-x4-upscaler"),
        **__hf_hub_vae_safetensors_dict__("stabilityai/stable-diffusion-x4-upscaler"),
    },
    "stable-v2": {
        **__hf_hub_stable_v2_dict__("stabilityai/stable-diffusion-2"),
        **__hf_hub_vae_dict("stabilityai/stable-diffusion-2"),
    },
    "stable-v2.1": {
        **__hf_hub_stable_v2_1_dict__("stabilityai/stable-diffusion-2-1"),
        **__hf_hub_vae_dict("stabilityai/stable-diffusion-2-1"),
    },
    "stable-v2.1-inpainting": {
        **__hf_hub_stable_v2_1_dict__("stabilityai/stable-diffusion-2-inpainting"),
        **__hf_hub_vae_dict("stabilityai/stable-diffusion-2-inpainting"),
    },
    "stable-xl-base": {
        **__hf_hub_stable_xl_safetensors_dict__(
            "stabilityai/stable-diffusion-xl-base-1.0"
        ),
        **__hf_hub_vae_safetensors_dict__("stabilityai/stable-diffusion-xl-base-1.0"),
    },
    "stable-xl-base-vae-fp16": {
        **__hf_hub_stable_xl_safetensors_dict__(
            "stabilityai/stable-diffusion-xl-base-1.0"
        ),
        **__hf_hub_vae_safetensors_dict__("madebyollin/sdxl-vae-fp16-fix"),
    },
    "stable-xl-turbo": {
        **__hf_hub_stable_xl_safetensors_dict__("stabilityai/sdxl-turbo"),
        **__hf_hub_vae_safetensors_dict__("stabilityai/sdxl-turbo"),
    },
    "stable-xl-realism-engine-v30": {
        **__hf_hub_stable_xl_safetensors_dict__("misri/realismEngineSDXL_v30VAE"),
        **__hf_hub_vae_safetensors_dict__("misri/realismEngineSDXL_v30VAE"),
    },
    "stable-xl-base-inpainting-0.1": {
        **__hf_hub_stable_xl_safetensors_dict__(
            "diffusers/stable-diffusion-xl-1.0-inpainting-0.1"
        ),
        **__hf_hub_vae_safetensors_fp16_dict__(
            "diffusers/stable-diffusion-xl-1.0-inpainting-0.1"
        ),
    },
    "stable-xl-opendalle-v1.1": {
        **__hf_hub_stable_xl_safetensors_dict__("dataautogpt3/OpenDalleV1.1"),
        **__hf_hub_vae_safetensors_dict__("dataautogpt3/OpenDalleV1.1"),
    },
    "stable-xl-realvis-v4.0": {
        **__hf_hub_stable_xl_safetensors_dict__("SG161222/RealVisXL_V4.0"),
        **__hf_hub_vae_safetensors_dict__("SG161222/RealVisXL_V4.0"),
    },
    "stable-xl-realvis-v4.0-inpainting": {
        **__hf_hub_stable_xl_safetensors_dict__("OzzyGT/RealVisXL_V4.0_inpainting"),
        **__hf_hub_vae_safetensors_fp16_dict__("OzzyGT/RealVisXL_V4.0_inpainting"),
    },
    "stable-xl-juggernaut-v8": {
        **__hf_hub_stable_xl_dict__("RunDiffusion/Juggernaut-XL-v8"),
        **__hf_hub_vae_dict("RunDiffusion/Juggernaut-XL-v8"),
    },
    "stable-xl-playground-v2-aesthetic": {
        **__hf_hub_stable_xl_safetensors_dict__(
            "playgroundai/playground-v2-1024px-aesthetic"
        ),
        **__hf_hub_vae_safetensors_dict__(
            "playgroundai/playground-v2-1024px-aesthetic"
        ),
    },
    "stable-v3-medium": {
        **__hf_hub_stable_3_safetensors_dict__(
            "ckpt/stable-diffusion-3-medium-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("ckpt/stable-diffusion-3-medium-diffusers"),
    },
    "stable-v3.5-medium": {
        **__hf_hub_stable_3_safetensors_dict__("ckpt/stable-diffusion-3.5-medium"),
        **__hf_hub_vae_safetensors_dict__("ckpt/stable-diffusion-3.5-medium"),
    },
    "stable-v3.5-large": {
        **__hf_hub_stable_3_safetensors_dict__("yuvraj108c/stable-diffusion-3.5-large"),
        **__hf_hub_vae_safetensors_dict__("yuvraj108c/stable-diffusion-3.5-large"),
    },
    "stable-flux-schnell": {
        **__hf_hub_stable_flux_safetensors_dict__("lzyvegetable/FLUX.1-schnell"),
        **__hf_hub_vae_safetensors_dict__("lzyvegetable/FLUX.1-schnell"),
    },
    "stable-flux-dev": {
        **__hf_hub_stable_flux_safetensors_dict__("camenduru/FLUX.1-dev-diffusers"),
        **__hf_hub_vae_safetensors_dict__("camenduru/FLUX.1-dev-diffusers"),
    },
    "stable-flux-dev-fill": {
        **__hf_hub_stable_flux_safetensors_dict__(
            "fuliucansheng/FLUX.1-Fill-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Fill-dev-diffusers"),
    },
    "stable-flux-dev-canny": {
        **__hf_hub_stable_flux_ctrl_safetensors_dict__(
            "fuliucansheng/FLUX.1-Canny-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Canny-dev-diffusers"),
    },
    "stable-flux-dev-depth": {
        **__hf_hub_stable_flux_ctrl_safetensors_dict__(
            "fuliucansheng/FLUX.1-Depth-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Depth-dev-diffusers"),
    },
    "stable-flux-dev-redux": {
        **__hf_hub_stable_flux_safetensors_dict__("camenduru/FLUX.1-dev-diffusers"),
        **__hf_hub_vae_safetensors_dict__("camenduru/FLUX.1-dev-diffusers"),
        **__hf_flux_dev_redux_safetensors_dict__("tentpole/flux1-dev-redux"),
    },
    "stable-flux-dev-redux-fill": {
        **__hf_hub_stable_flux_safetensors_dict__(
            "fuliucansheng/FLUX.1-Fill-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Fill-dev-diffusers"),
        **__hf_flux_dev_redux_safetensors_dict__("tentpole/flux1-dev-redux"),
    },
    "stable-flux-dev-redux-canny": {
        **__hf_hub_stable_flux_ctrl_safetensors_dict__(
            "fuliucansheng/FLUX.1-Canny-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Canny-dev-diffusers"),
        **__hf_flux_dev_redux_safetensors_dict__("tentpole/flux1-dev-redux"),
    },
    "stable-flux-dev-redux-depth": {
        **__hf_hub_stable_flux_ctrl_safetensors_dict__(
            "fuliucansheng/FLUX.1-Depth-dev-diffusers"
        ),
        **__hf_hub_vae_safetensors_dict__("fuliucansheng/FLUX.1-Depth-dev-diffusers"),
        **__hf_flux_dev_redux_safetensors_dict__("tentpole/flux1-dev-redux"),
    },
    "stable-video-diffusion-img2vid-xt": {
        **__hf_hub_stable_video_safetensors_dict__(
            "stabilityai/stable-video-diffusion-img2vid-xt"
        ),
        **__hf_hub_vae_safetensors_dict__(
            "stabilityai/stable-video-diffusion-img2vid-xt"
        ),
    },
    "stable-video-diffusion-img2vid-xt-1-1": {
        **__hf_hub_stable_video_safetensors_dict__(
            "vdo/stable-video-diffusion-img2vid-xt-1-1"
        ),
        **__hf_hub_vae_safetensors_dict__("vdo/stable-video-diffusion-img2vid-xt-1-1"),
    },
    "wan-v2.1-t2v-1.3b": {
        **__hf_hub_wan_v2_1_safetensors_dict__("Wan-AI/Wan2.1-T2V-1.3B-Diffusers"),
        **__hf_hub_vae_safetensors_dict__("Wan-AI/Wan2.1-T2V-1.3B-Diffusers"),
    },
    "wan-v2.1-t2v-14b": {
        **__hf_hub_wan_v2_1_safetensors_dict__(
            "Wan-AI/Wan2.1-T2V-14B-Diffusers", n1=12
        ),
        **__hf_hub_vae_safetensors_dict__("Wan-AI/Wan2.1-T2V-14B-Diffusers"),
    },
    "wan-v2.1-i2v-14b-480p": {
        **__hf_hub_wan_v2_1_safetensors_dict__(
            "Wan-AI/Wan2.1-I2V-14B-480P-Diffusers", n1=14, im=True
        ),
        **__hf_hub_vae_safetensors_dict__("Wan-AI/Wan2.1-I2V-14B-480P-Diffusers"),
    },
    "wan-v2.1-i2v-14b-720p": {
        **__hf_hub_wan_v2_1_safetensors_dict__(
            "Wan-AI/Wan2.1-I2V-14B-720P-Diffusers", n1=14, im=True
        ),
        **__hf_hub_vae_safetensors_dict__("Wan-AI/Wan2.1-I2V-14B-720P-Diffusers"),
    },
}

pretrained_stable_extensions_infos = {
    # sd 1.5 controlnet
    "stable-v1.5-controlnet-canny": __hf_hub_controlnet_dict__(
        "lllyasviel/sd-controlnet-canny"
    ),
    "stable-v1.5-controlnet-openpose": __hf_hub_controlnet_dict__(
        "lllyasviel/sd-controlnet-openpose"
    ),
    "stable-v1.5-controlnet-softedge": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_softedge",
    ),
    "stable-v1.5-controlnet-inpainting": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_inpaint"
    ),
    "stable-v1.5-controlnet-depth": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11f1p_sd15_depth"
    ),
    "stable-v1.5-controlnet-tile": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11f1e_sd15_tile"
    ),
    "stable-v1.5-controlnet-lineart": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_lineart"
    ),
    "stable-v1.5-controlnet-softedge": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_softedge"
    ),
    "stable-v1.5-controlnet-scribble": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_scribble"
    ),
    "stable-v1.5-controlnet-openpose": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_openpose"
    ),
    # sd 1.5 adapter
    "stable-v1.5-adapter-canny": __hf_hub_adapter_dict__(
        "diffusers/t2iadapter_canny_sd15v2"
    ),
    "stable-v1.5-adapter-depth": __hf_hub_adapter_dict__(
        "diffusers/t2iadapter_depth_sd15v2"
    ),
    "stable-v1.5-adapter-sketch": __hf_hub_adapter_dict__(
        "diffusers/t2iadapter_sketch_sd15v2"
    ),
    # sdxl controlnet
    "stable-xl-controlnet-canny": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-canny-sdxl-1.0"
    ),
    "stable-xl-controlnet-softedge-dexined": __hf_hub_controlnet_dict__(
        "SargeZT/controlnet-sd-xl-1.0-softedge-dexined"
    ),
    "stable-xl-controlnet-depth": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-depth-sdxl-1.0"
    ),
    "stable-xl-controlnet-depth-small": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-depth-sdxl-1.0-small"
    ),
    "stable-xl-controlnet-tile": __hf_hub_controlnet_safetensors_dict__(
        "xinsir/controlnet-tile-sdxl-1.0"
    ),
    "stable-xl-controlnet-openpose": __hf_hub_controlnet_safetensors_dict__(
        "xinsir/controlnet-openpose-sdxl-1.0"
    ),
    "stable-xl-controlnet-scribble": __hf_hub_controlnet_safetensors_dict__(
        "xinsir/controlnet-scribble-sdxl-1.0"
    ),
    # sdxl adapter
    "stable-xl-adapter-canny": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-canny-sdxl-1.0"
    ),
    "stable-xl-adapter-sketch": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-sketch-sdxl-1.0"
    ),
    "stable-xl-adapter-depth": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-depth-zoe-sdxl-1.0"
    ),
    "stable-xl-adapter-depth-midas": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-depth-midas-sdxl-1.0"
    ),
    "stable-xl-adapter-openpose": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-openpose-sdxl-1.0"
    ),
    "stable-xl-adapter-lineart": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-lineart-sdxl-1.0"
    ),
    # sdxl refiner
    "stable-xl-refiner-1.0": {
        "refiner": {
            "unet": {
                "config": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/unet/config.json"
                ),
                "weight": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/unet/diffusion_pytorch_model.fp16.safetensors"
                ),
            },
            "text2": {
                "config": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/text_encoder_2/config.json"
                ),
                "vocab": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/tokenizer_2/vocab.json"
                ),
                "merge": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/tokenizer_2/merges.txt"
                ),
                "weight": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/text_encoder_2/model.fp16.safetensors"
                ),
            },
            "vae": {
                "config": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/vae/config.json"
                ),
                "weight": hf_endpoint_url(
                    f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/vae/diffusion_pytorch_model.fp16.safetensors"
                ),
            },
            "scheduler": hf_endpoint_url(
                f"/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/scheduler/scheduler_config.json"
            ),
        },
    },
    # sdxl vae
    "stable-xl-vae-fp16": {
        "vae": {
            "config": hf_endpoint_url(
                "/madebyollin/sdxl-vae-fp16-fix/resolve/main/config.json"
            ),
            "weight": hf_endpoint_url(
                "/madebyollin/sdxl-vae-fp16-fix/resolve/main/diffusion_pytorch_model.safetensors"
            ),
        },
    },
    # stable 3 controlnet
    "stable-v3-controlnet-canny": __hf_hub_controlnet_safetensors_dict__(
        "InstantX/SD3-Controlnet-Canny"
    ),
    "stable-v3-controlnet-tile": __hf_hub_controlnet_safetensors_dict__(
        "InstantX/SD3-Controlnet-Tile"
    ),
    # stable flux controlnet
    "stable-flux-controlnet-dev-union": __hf_hub_controlnet_safetensors_dict__(
        "InstantX/FLUX.1-dev-Controlnet-Union"
    ),
    "stable-flux-controlnet-dev-union-pro": __hf_hub_controlnet_safetensors_dict__(
        "Shakker-Labs/FLUX.1-dev-ControlNet-Union-Pro"
    ),
}

from unitorch.cli.models.diffusion_utils import load_weight

from unitorch.cli.models.diffusers.modeling_stable import (
    StableForText2ImageGeneration,
    StableForImage2ImageGeneration,
    StableForImageInpainting,
    StableForImageResolution,
)
from unitorch.cli.models.diffusers.modeling_stable_xl import (
    StableXLForText2ImageGeneration,
    StableXLForImage2ImageGeneration,
    StableXLForImageInpainting,
)

from unitorch.cli.models.diffusers.modeling_stable_3 import (
    Stable3ForText2ImageGeneration,
    Stable3ForImage2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_stable_flux import (
    StableFluxForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_controlnet import (
    ControlNetForText2ImageGeneration,
    ControlNetForImage2ImageGeneration,
    ControlNetForImageInpainting,
)
from unitorch.cli.models.diffusers.modeling_controlnet_xl import (
    ControlNetXLForText2ImageGeneration,
    ControlNetXLForImage2ImageGeneration,
    ControlNetXLForImageInpainting,
)

from unitorch.cli.models.diffusers.modeling_controlnet_3 import (
    ControlNet3ForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_controlnet_flux import (
    ControlNetFluxForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_adapter import (
    StableAdapterForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_adapter_xl import (
    StableXLAdapterForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_wan import (
    WanForText2VideoGeneration,
    WanForImage2VideoGeneration,
)
from unitorch.cli.models.diffusers.modeling_vae import VAEForDiffusion
from unitorch.cli.models.diffusers.processing_stable import StableProcessor
from unitorch.cli.models.diffusers.processing_stable_xl import StableXLProcessor
from unitorch.cli.models.diffusers.processing_stable_3 import Stable3Processor
from unitorch.cli.models.diffusers.processing_stable_flux import StableFluxProcessor
from unitorch.cli.models.diffusers.processing_controlnet import ControlNetProcessor
from unitorch.cli.models.diffusers.processing_controlnet_xl import ControlNetXLProcessor
from unitorch.cli.models.diffusers.processing_controlnet_3 import ControlNet3Processor
from unitorch.cli.models.diffusers.processing_controlnet_flux import (
    ControlNetFluxProcessor,
)
from unitorch.cli.models.diffusers.processing_adapter import AdapterProcessor
from unitorch.cli.models.diffusers.processing_adapter_xl import AdapterXLProcessor

if is_opencv_available():
    from unitorch.cli.models.diffusers.processing_wan import WanProcessor



# Copyright 2023 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""PyTorch implementation of the Lion optimizer."""
import torch
from torch.optim.optimizer import Optimizer


class Lion(Optimizer):
    r"""Implements Lion algorithm."""

    def __init__(self, params, lr=1e-4, betas=(0.9, 0.99), weight_decay=0.0):
        """Initialize the hyperparameters.
        Args:
          params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
          lr (float, optional): learning rate (default: 1e-4)
          betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.99))
          weight_decay (float, optional): weight decay coefficient (default: 0)
        """

        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, weight_decay=weight_decay)
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.
        Args:
          closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
        Returns:
          the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                # Perform stepweight decay
                p.data.mul_(1 - group["lr"] * group["weight_decay"])

                grad = p.grad
                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p)

                exp_avg = state["exp_avg"]
                beta1, beta2 = group["betas"]

                # Weight update
                update = exp_avg * beta1 + grad * (1 - beta1)
                p.add_(torch.sign(update), alpha=-group["lr"])
                # Decay the momentum running average coefficient
                exp_avg.mul_(beta2).add_(grad, alpha=1 - beta2)

        return loss



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.utils import is_diffusers_available
import unitorch.modules.replace.datasets_v2

if is_diffusers_available():
    import unitorch.modules.replace.diffusers_v2



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from PIL import Image
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.clip import ClipProcessor as _ClipProcessor
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import cached_path
from unitorch.cli import WriterOutputs
from unitorch.cli.models import (
    TensorsInputs,
    GenerationOutputs,
    GenerationTargets,
)
from unitorch.cli.models.clip import pretrained_clip_infos


class ClipProcessor(_ClipProcessor):
    """Processor for the CLIP model."""

    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 128,
        position_start_id: Optional[int] = 0,
    ):
        """
        Initialize the ClipProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            merge_path (str): The path to the BPE merge file.
            vision_config_path (str): The path to the vision configuration file.
            max_seq_length (int, optional): The maximum sequence length. Defaults to 128.
            position_start_id (int, optional): The position start ID. Defaults to 0.
        """
        super().__init__(
            vocab_path=vocab_path,
            merge_path=merge_path,
            vision_config_path=vision_config_path,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

    @classmethod
    @add_default_section_for_init("core/process/clip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ClipProcessor from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the processor's initialization arguments.
        """
        config.set_default_section("core/process/clip")
        pretrained_name = config.getoption("pretrained_name", "clip-vit-base-patch16")
        vocab_path = config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        merge_path = config.getoption("merge_path", None)
        merge_path = pop_value(
            merge_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "merge"),
        )
        merge_path = cached_path(merge_path)

        vision_config_path = config.getoption("vision_config_path", None)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vision_config"),
        )

        vision_config_path = cached_path(vision_config_path)

        return {
            "vocab_path": vocab_path,
            "merge_path": merge_path,
            "vision_config_path": vision_config_path,
        }

    @register_process("core/process/clip/classification")
    def _classification(
        self,
        text: str,
        image: Union[Image.Image, str],
        max_seq_length: Optional[int] = None,
    ):
        """
        Process text and image inputs for classification.

        Args:
            text (str): The input text.
            image (Union[Image.Image, str]): The input image.
            max_seq_length (int, optional): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The processed inputs as tensors.
        """
        if isinstance(image, str):
            image = Image.open(image)

        outputs = super().classification(
            text=text,
            image=image,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            position_ids=outputs.position_ids,
            pixel_values=outputs.pixel_values,
        )

    @register_process("core/process/clip/text_classification")
    def _text_classification(
        self,
        text: str,
        max_seq_length: Optional[int] = None,
    ):
        """
        Process text inputs for text classification.

        Args:
            text (str): The input text.
            max_seq_length (int, optional): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The processed inputs as tensors.
        """
        outputs = super().text_classification(
            text=text,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            position_ids=outputs.position_ids,
        )

    @register_process("core/process/clip/image_classification")
    def _image_classification(
        self,
        image: Union[Image.Image, str],
    ):
        """
        Process image inputs for image classification.

        Args:
            image (Union[Image.Image, str]): The input image.

        Returns:
            TensorsInputs: The processed inputs as tensors.
        """
        if isinstance(image, str):
            image = Image.open(image)
        outputs = super().image_classification(image=image)
        return TensorsInputs(pixel_values=outputs.pixel_values)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from PIL import Image
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from torchvision.transforms import Resize, CenterCrop, ToTensor, Normalize, Compose
from unitorch.models import GenericOutputs


class BRIAProcessor:
    def __init__(
        self,
        image_size: Optional[int] = 1024,
    ):
        self.image_size = image_size
        self.transform_inputs = Compose(
            [
                Resize(size=[self.image_size, self.image_size]),
                ToTensor(),
                Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        )
        self.transform_labels = Compose(
            [
                Resize(size=[self.image_size, self.image_size]),
                ToTensor(),
            ]
        )

    def segmentation_inputs(self, image: Image):
        width, height = image.size
        image = self.transform_inputs(image.convert("RGB"))
        return GenericOutputs(
            image=torch.tensor(image),
            sizes=torch.tensor([height, width]),
        )

    def segmentation_labels(self, image: Image):
        width, height = image.size
        image = self.transform_labels(image.convert("L"))
        return GenericOutputs(
            image=torch.tensor(image),
            sizes=torch.tensor([height, width]),
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import autocast
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from unitorch.models.bria import BRIAForSegmentation as _BRIAForSegmentation
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import (
    detection_model_decorator,
    segmentation_model_decorator,
    DetectionOutputs,
    SegmentationOutputs,
    LossOutputs,
)


@register_model("core/model/segmentation/bria", segmentation_model_decorator)
class BRIAForSegmentation(_BRIAForSegmentation):
    def __init__(self):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/model/segmentation/bria")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/segmentation/bria")

        inst = cls()
        weight_path = config.getoption("pretrained_weight_path", None)
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    def forward(self, images):
        outputs = super().forward(images)
        return SegmentationOutputs(
            masks=outputs,
        )

    def segment(self, images, sizes: Optional[List[Tuple[int, int]]] = None):
        outputs = super().forward(images)
        if sizes is None:
            return SegmentationOutputs(
                masks=outputs.logits,
            )
        masks = [
            F.interpolate(mask.unsqueeze(0), size=list(size), mode="bilinear").squeeze(
                0
            )
            for mask, size in zip(outputs.logits, sizes)
        ]
        masks = [m.permute(1, 2, 0) for m in masks]
        masks = [(m - m.min()) / (m.max() - m.min()) for m in masks]
        return SegmentationOutputs(
            masks=masks,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import CLIPTokenizer, CLIPImageProcessor
from torchvision.transforms import (
    Resize,
    CenterCrop,
    RandomCrop,
    Lambda,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from diffusers.image_processor import VaeImageProcessor
from unitorch.models import (
    HfTextClassificationProcessor,
    HfImageClassificationProcessor,
    GenericOutputs,
)


class StableProcessor(HfTextClassificationProcessor):
    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        vae_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token: Optional[str] = "<|endoftext|>",
        image_size: Optional[Tuple[int, int]] = None,
        low_res_image_size: Optional[Tuple[int, int]] = None,
        center_crop: Optional[bool] = False,
        random_flip: Optional[bool] = False,
    ):
        tokenizer = CLIPTokenizer(
            vocab_file=vocab_path,
            merges_file=merge_path,
        )

        tokenizer.cls_token = tokenizer.bos_token
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.pad_token = pad_token

        HfTextClassificationProcessor.__init__(
            self,
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        if image_size is not None:
            self.image_size = (
                image_size
                if isinstance(image_size, tuple)
                else (image_size, image_size)
            )
        else:
            self.image_size = None
        if self.image_size is not None:
            self.vision_processor = Compose(
                [
                    Resize((self.image_size[1], self.image_size[0])),
                    CenterCrop((self.image_size[1], self.image_size[0]))
                    if center_crop
                    else RandomCrop((self.image_size[1], self.image_size[0])),
                    RandomHorizontalFlip() if random_flip else Lambda(lambda x: x),
                    ToTensor(),
                    Normalize([0.5], [0.5]),
                ]
            )
            self.condition_vision_processor = Compose(
                [
                    Resize((self.image_size[1], self.image_size[0])),
                    CenterCrop((self.image_size[1], self.image_size[0])),
                    ToTensor(),
                ]
            )
        else:
            self.vision_processor = Compose(
                [
                    RandomHorizontalFlip() if random_flip else Lambda(lambda x: x),
                    ToTensor(),
                    Normalize([0.5], [0.5]),
                ]
            )
            self.condition_vision_processor = Compose(
                [
                    ToTensor(),
                ]
            )

        if vae_config_path is not None:
            vae_config_dict = json.load(open(vae_config_path))
            vae_scale_factor = 2 ** (
                len(vae_config_dict.get("block_out_channels", [])) - 1
            )
            self.vae_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor
            )
            self.vae_condition_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor,
                do_convert_rgb=True,
                do_normalize=False,
            )
        else:
            self.vae_image_processor = None
            self.vae_condition_image_processor = None
        self.divisor = 8
        self.low_res_image_size = low_res_image_size

    def text2image(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)

        image = image.convert("RGB")

        pixel_values = self.vision_processor(image)

        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)

        return GenericOutputs(
            pixel_values=pixel_values,
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
        )

    def text2image_inputs(
        self,
        prompt: str,
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)
        negative_prompt_outputs = self.classification(
            negative_prompt, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
        )

    def image2image_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        size = image.size if self.image_size is None else self.image_size
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def inpainting_inputs(
        self,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")

        size = image.size if self.image_size is None else self.image_size
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)
        mask_image = mask_image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_image_processor.preprocess(mask_image)[0]
        pixel_masks = (pixel_masks + 1) / 2

        return GenericOutputs(
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
        )

    def resolution_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        size = (
            image.size if self.low_res_image_size is None else self.low_res_image_size
        )
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def controlnet_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        size = image.size if self.image_size is None else self.image_size
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_condition_image_processor.preprocess(image)[0]
        return GenericOutputs(pixel_values=pixel_values)

    def controlnets_inputs(
        self,
        images: List[Union[Image.Image, str]],
    ):
        pixel_values = []
        for image in images:
            if isinstance(image, str):
                image = Image.open(image)
            image = image.convert("RGB")
            size = image.size if self.image_size is None else self.image_size
            size = (
                size[0] // self.divisor * self.divisor,
                size[1] // self.divisor * self.divisor,
            )
            image = image.resize(size, resample=Image.LANCZOS)

            pixel_values.append(self.vae_condition_image_processor.preprocess(image)[0])

        return GenericOutputs(pixel_values=torch.stack(pixel_values, dim=0))

    def inpainting_control_inputs(
        self,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        size = image.size if self.image_size is None else self.image_size
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")
        mask_image = mask_image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_image_processor.preprocess(mask_image)[0]
        pixel_values[:, pixel_masks[0] > 0.5] = -1.0
        return GenericOutputs(pixel_values=pixel_values)

    def adapter_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        size = image.size if self.image_size is None else self.image_size
        size = (
            size[0] // self.divisor * self.divisor,
            size[1] // self.divisor * self.divisor,
        )
        image = image.resize(size, resample=Image.LANCZOS)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def adapters_inputs(
        self,
        images: List[Union[Image.Image, str]],
    ):
        pixel_values = []
        for image in images:
            if isinstance(image, str):
                image = Image.open(image)
            image = image.convert("RGB")
            size = image.size if self.image_size is None else self.image_size
            size = (
                size[0] // self.divisor * self.divisor,
                size[1] // self.divisor * self.divisor,
            )
            image = image.resize(size, resample=Image.LANCZOS)

            pixel_values.append(self.vae_image_processor.preprocess(image)[0])

        return GenericOutputs(pixel_values=torch.stack(pixel_values, dim=0))


class StableVideoProcessor(HfImageClassificationProcessor):
    def __init__(
        self,
        vision_config_path: str,
        vae_config_path: str,
    ):
        vision_processor = CLIPImageProcessor.from_json_file(vision_config_path)
        super().__init__(
            vision_processor=vision_processor,
        )
        vae_config_dict = json.load(open(vae_config_path))
        vae_scale_factor = 2 ** (len(vae_config_dict.get("block_out_channels", [])) - 1)
        self.vae_image_processor = VaeImageProcessor(vae_scale_factor=vae_scale_factor)

    def image2video_inputs(
        self,
        image: Union[Image.Image, str],
        vae_image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.resize((224, 224), resample=Image.LANCZOS)
        pixel_outputs = self.classification(image=image)
        if isinstance(vae_image, str):
            vae_image = Image.open(vae_image)
        vae_pixel_values = self.vae_image_processor.preprocess(vae_image)[0]

        return GenericOutputs(
            pixel_values=pixel_outputs.pixel_values,
            vae_pixel_values=vae_pixel_values,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import io
import torch
import hashlib
import logging
import requests
import numpy as np
import torch.nn as nn
from PIL import Image
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import numpy_to_pil
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli.models.modeling_utils import ListTensorsOutputs, ListTensorsTargets


@dataclass
class SegmentationOutputs(ListTensorsOutputs, WriterMixin):
    masks: Union[torch.Tensor, List[torch.Tensor]]
    classes: Union[torch.Tensor, List[torch.Tensor]] = torch.empty(0)


@dataclass
class SegmentationTargets(ListTensorsTargets):
    targets: Union[torch.Tensor, List[torch.Tensor]]
    sample_weight: Optional[torch.Tensor] = torch.empty(0)


class SegmentationProcessor:
    def __init__(
        self,
        mask_threshold: float = None,
        output_folder: Optional[str] = None,
        http_url: Optional[str] = None,
    ):
        self.mask_threshold = mask_threshold

        self.output_folder = output_folder
        if self.output_folder is not None and not os.path.exists(output_folder):
            os.makedirs(self.output_folder, exist_ok=True)
        self.http_url = http_url

    @classmethod
    @add_default_section_for_init("core/process/segmentation")
    def from_core_configure(cls, config, **kwargs):
        pass

    def save_image(self, image: Image.Image):
        md5 = hashlib.md5()
        md5.update(image.tobytes())
        name = md5.hexdigest() + ".jpg"
        if self.http_url is not None:
            byte = io.BytesIO()
            image.save(byte, format="JPEG")
            for _ in range(3):
                resp = requests.post(
                    self.http_url.format(name), files={"file": byte.getvalue()}
                )
                if resp.status_code == 200:
                    break
            if resp.status_code != 200:
                logging.error(f"Failed to save image {name} to zip.")
        else:
            output_path = f"{self.output_folder}/{name}"
            image.save(output_path)
        return name

    @register_process("core/postprocess/segmentation")
    def _segmentation(
        self,
        outputs: SegmentationOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == len(outputs.masks)
        mask0 = outputs.masks[0].numpy()
        if (
            mask0.shape[-1] == 1 or mask0.ndim == 2
        ) and self.mask_threshold is not None:
            results["pixel_class"] = [
                (m.numpy() > self.mask_threshold).astype(np.int16).reshape(-1).tolist()
                for m in outputs.masks
            ]
        else:
            results["pixel_class"] = [
                m.numpy().argmax(-1).reshape(-1).tolist() for m in outputs.masks
            ]
        return WriterOutputs(results)

    @register_process("core/postprocess/segmentation/mask")
    def _segmentation_mask(
        self,
        outputs: SegmentationOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == len(outputs.masks)
        if self.mask_threshold is not None:
            masks = [
                (m.numpy() < self.mask_threshold).astype(np.int16)
                for m in outputs.masks
            ]
        else:
            masks = [m.numpy() for m in outputs.masks]
        results["mask_image"] = [
            self.save_image(numpy_to_pil(mask if mask.shape[0] > 1 else mask[0]))
            for mask in masks
        ]
        return WriterOutputs(results)

    @register_process("core/postprocess/segmentation/class_mask")
    def _segmentation_class_mask(
        self,
        outputs: SegmentationOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == len(outputs.masks)
        assert all([m.ndim == 3 for m in outputs.masks])
        if self.mask_threshold is not None:
            masks = [
                (m.numpy() < self.mask_threshold).astype(np.int16)
                for m in outputs.masks
            ]
        else:
            masks = [m.numpy() for m in outputs.masks]
        classes = [c.numpy() for c in outputs.classes]
        classes = [c if c.ndim == 1 else c.argmax(-1) for c in classes]
        results["mask_images"] = [
            ";".join(
                [
                    self.save_image(numpy_to_pil(_mask_image))
                    for _mask_image in _mask_images
                ]
            )
            for _mask_images in masks
        ]
        results["mask_classes"] = [
            ";".join([str(_class) for _class in _classes]) for _classes in classes
        ]
        return WriterOutputs(results)


def segmentation_model_decorator(cls):
    class SegmentationModel(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            if "__segmentation_model__" in kwargs:
                self.model = kwargs.pop("__segmentation_model__")
            else:
                self.model = cls(*args, **kwargs)

            __more_attrs__ = [
                "load_state_dict",
                "state_dict",
                "save_checkpoint",
                "from_checkpoint",
                "from_pretrained",
            ]
            for __more_attr__ in __more_attrs__:
                setattr(self, __more_attr__, getattr(self.model, __more_attr__))

            assert hasattr(self.model, "segment")

        def forward(self, *args, **kwargs):
            if self.training:
                return self.model(*args, **kwargs)
            return self.model.segment(*args, **kwargs)

        @classmethod
        def from_core_configure(_cls, cfg, **kwargs):
            model = cls.from_core_configure(cfg, **kwargs)
            return _cls(__segmentation_model__=model)

    return SegmentationModel



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_vit_infos = {
    "vit-base-patch16-224-in21k": {
        "config": hf_endpoint_url(
            "/google/vit-base-patch16-224-in21k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-base-patch16-224-in21k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-base-patch16-224-in21k/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-base-patch32-224-in21k": {
        "config": hf_endpoint_url(
            "/google/vit-base-patch32-224-in21k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-base-patch32-224-in21k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-base-patch32-224-in21k/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-large-patch16-224-in21k": {
        "config": hf_endpoint_url(
            "/google/vit-large-patch16-224-in21k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-large-patch16-224-in21k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-large-patch16-224-in21k/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-large-patch32-224-in21k": {
        "config": hf_endpoint_url(
            "/google/vit-large-patch32-224-in21k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-large-patch32-224-in21k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-large-patch32-224-in21k/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-huge-patch14-224-in21k": {
        "config": hf_endpoint_url(
            "/google/vit-huge-patch14-224-in21k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-huge-patch14-224-in21k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-huge-patch14-224-in21k/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-base-patch16-224": {
        "config": hf_endpoint_url(
            "/google/vit-base-patch16-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-base-patch16-224/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-base-patch16-224/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-base-patch16-384": {
        "config": hf_endpoint_url(
            "/google/vit-base-patch16-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-base-patch16-384/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-base-patch16-384/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-base-patch32-384": {
        "config": hf_endpoint_url(
            "/google/vit-base-patch32-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-base-patch32-384/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-base-patch32-384/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-large-patch16-224": {
        "config": hf_endpoint_url(
            "/google/vit-large-patch16-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-large-patch16-224/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-large-patch16-224/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-large-patch16-384": {
        "config": hf_endpoint_url(
            "/google/vit-large-patch16-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-large-patch16-384/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-large-patch16-384/resolve/main/pytorch_model.bin"
        ),
    },
    "vit-large-patch32-384": {
        "config": hf_endpoint_url(
            "/google/vit-large-patch32-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/vit-large-patch32-384/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/google/vit-large-patch32-384/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.vit.modeling
import unitorch.cli.models.vit.processing
from unitorch.cli.models.vit.modeling import ViTForImageClassification
from unitorch.cli.models.vit.processing import ViTProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import time
import json
import logging
import numpy as np
import pandas as pd
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
from PIL import Image
from itertools import chain
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, Iterator
from torch.utils.data import DataLoader, Dataset, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from torch.cuda.amp import GradScaler
from torch.multiprocessing import Process, Queue
from unitorch import set_seed
from unitorch.models import ExponentialMovingAverage
from unitorch.utils import get_local_rank
from unitorch.utils import (
    DistributedSkipSampler,
    RandomSkipSampler,
    SequentialSkipSampler,
    PostProcess,
    IOProcess,
    GENERATE_FINISHED,
)
from unitorch.models import GenericOutputs
from unitorch.cli import (
    cached_path,
    register_task,
    registered_model,
    registered_optim,
    registered_dataset,
    registered_loss,
    registered_score,
    registered_scheduler,
    registered_writer,
    init_registered_module,
    init_registered_process,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models import (
    ModelInputs,
    ModelOutputs,
    ModelTargets,
    LossOutputs,
    CombineTensorsInputs,
    CombineTensorsTargets,
)
import unitorch.cli.wandb as wandb


class DatasetFeature(Dataset):
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, idx):
        row = self.dataset[idx]
        ret = {}
        for k, v in row.items():
            if isinstance(v, Image.Image):
                v = np.array(v).tolist()
            if not isinstance(v, str):
                v = json.dumps(v)
            ret[k] = v
        return ret

    def __len__(self):
        return len(self.dataset)


def collate_fn(bufs):
    combine_inputs, combine_targets = list(zip(*bufs))
    if isinstance(combine_inputs[0], ModelInputs):
        inputs = type(combine_inputs[0]).stack(*combine_inputs)
    else:
        combine_inputs = [
            type(_inputs[0]).stack(*_inputs) for _inputs in list(zip(*combine_inputs))
        ]
        inputs = CombineTensorsInputs()
        for _inputs in combine_inputs:
            inputs.add(_inputs)

    if isinstance(combine_targets[0], ModelTargets):
        targets = type(combine_targets[0]).stack(*combine_targets)
    else:
        combine_targets = [
            type(_targets[0]).stack(*_targets)
            for _targets in list(zip(*combine_targets))
        ]
        targets = CombineTensorsTargets()
        for _targets in combine_targets:
            targets.add(_targets)

    return inputs, targets


@torch.no_grad()
def infer(model, iter_data):
    model.eval()
    outputs, targets = [], []
    for _, (_inputs, _targets) in enumerate(iter_data):
        if torch.cuda.is_available():
            _inputs = _inputs.cuda()
            _targets = _targets.cuda()
        _outputs = model(**_inputs.dict())
        outputs.append(_outputs)
        targets.append(_targets)

    if isinstance(outputs[0], LossOutputs):
        outputs = LossOutputs(
            loss=torch.tensor([output.loss for output in outputs]).to(
                device=outputs[0].loss.device
            )
        )
        if dist.is_initialized():
            outputs = outputs.cuda().sync().cpu()
        else:
            outputs = outputs.cpu()
    else:
        outputs = type(outputs[0]).union(*outputs)
        targets = type(targets[0]).union(*targets)

        if dist.is_initialized():
            outputs = outputs.cuda().sync().cpu()
            targets = targets.cuda().sync().cpu()
        else:
            outputs = outputs.cpu()
            targets = targets.cpu()
    model.train()
    return GenericOutputs(outputs=outputs, targets=targets)


def monitor(outputs, targets, monitor_fns):
    if monitor_fns is None:
        return

    for monitor_fn in monitor_fns:
        score = monitor_fn(outputs=outputs, targets=targets)
        info = str(type(monitor_fn).__name__)
        logging.info(f"{info} is {score}")
        if wandb.is_available():
            wandb.log(
                {
                    f"val/{info}": score,
                }
            )
    return


def save_snapshot(
    model,
    ckpt_dir,
    iter_dev,
    score_fn,
    monitor_fns,
    optim,
    scheduler,
    save_checkpoint="default",
    ema_model=None,
    best_score=-np.inf,
    info_path=None,
    local_rank=-1,
    **kwargs,
):
    if not os.path.exists(ckpt_dir):
        os.makedirs(ckpt_dir, exist_ok=True)

    if ema_model is not None:
        base_model = ema_model
    else:
        base_model = model

    snapshot_time = time.strftime("%Y%m%d_%H%M", time.localtime())
    results = infer(base_model, iter_dev)
    if local_rank in [-1, 0]:
        new_score = score_fn(outputs=results.outputs, targets=results.targets)
        monitor(results.outputs, results.targets, monitor_fns)
        if save_checkpoint in ["all", "default", "best"] and new_score > best_score:
            best_score = new_score
            if model:
                model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name="pytorch_model.bin",
                )
            if ema_model:
                ema_model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name="pytorch_ema_model.bin",
                )
            if optim:
                optim.save_checkpoint(
                    ckpt_dir=ckpt_dir, weight_name="pytorch_optim.bin"
                )
            if scheduler:
                scheduler.save_checkpoint(
                    ckpt_dir=ckpt_dir, weight_name="pytorch_scheduler.bin"
                )

        if save_checkpoint in ["all", "default", "latest"]:
            if model:
                model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name="pytorch_model_latest.bin",
                )
            if ema_model:
                ema_model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name="pytorch_ema_model_latest.bin",
                )
                kwargs["num_ema_steps"] = ema_model.num_steps
            if optim:
                optim.save_checkpoint(
                    ckpt_dir=ckpt_dir, weight_name="pytorch_optim_latest.bin"
                )
            if scheduler:
                scheduler.save_checkpoint(
                    ckpt_dir=ckpt_dir, weight_name="pytorch_scheduler_latest.bin"
                )

        if save_checkpoint in ["all", "every"]:
            if model:
                model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name=f"pytorch_model_{snapshot_time}.bin",
                )
            if ema_model:
                ema_model.save_checkpoint(
                    ckpt_dir=ckpt_dir,
                    weight_name=f"pytorch_ema_model_{snapshot_time}.bin",
                )
        if info_path is not None:
            json.dump({"best_score": best_score, **kwargs}, open(info_path, "w"))
    return best_score


@register_task("core/task/supervised")
class SupervisedTask:
    def __init__(
        self,
        configure,
        model,
        datasets,
        local_rank: Optional[int] = -1,
        seed: Optional[int] = 1123,
        cpu_offload: Optional[bool] = False,
    ):
        """
        Initialize the SupervisedTask.

        Args:
            configure: The configuration object.
            model: The model for the task.
            datasets: The datasets for training and evaluation.
            local_rank (optional): The local rank for distributed training. Defaults to -1.
            seed (optional): The random seed. Defaults to 1123.
        """
        set_seed(seed)
        self.n_gpu = 1 if torch.cuda.is_available() else 0
        if dist.is_initialized():
            self.n_gpu = dist.get_world_size()

        self.config = configure
        self.model = model
        self.datasets = datasets
        self.local_rank = local_rank

        if self.local_rank != -1:
            torch.cuda.set_device(self.local_rank)

        if torch.cuda.is_available() and not cpu_offload:
            self.model = self.model.cuda()

        self.best_score = -np.inf

    @classmethod
    @add_default_section_for_init("core/task/supervised")
    def from_core_configure(cls, config, **kwargs):
        """
        Create a SupervisedTask instance from the core configuration.

        Args:
            config: The core configuration object.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary containing the configuration, model, datasets, and local rank.
        """
        try:
            torch.distributed.init_process_group(backend="nccl", init_method="env://")
        except:
            logging.info("PyTorch is not in distributed mode")

        config.set_default_section("core/task/supervised")

        model = config.getoption("model", None)
        dataset = config.getoption("dataset", None)

        if model is not None:
            model = init_registered_module(model, config, registered_model)

        if dataset is not None:
            dataset = init_registered_module(dataset, config, registered_dataset)

        local_rank = config.getdefault(
            "core/cli",
            "local_rank",
            get_local_rank(),
        )
        cpu_offload = config.getoption("cpu_offload", False)

        return dict(
            configure=config,
            model=model,
            datasets=dataset,
            local_rank=local_rank,
            cpu_offload=cpu_offload,
        )

    @add_default_section_for_function("core/task/supervised")
    def train(
        self,
        optim: str,
        loss_fn: str,
        score_fn: str,
        monitor_fns: Optional[Union[str, List[str]]] = None,
        scheduler: Optional[str] = None,
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        to_ckpt_dir: Optional[str] = "./to_ckpt",
        train_batch_size: Optional[int] = 128,
        dev_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        save_optimizer: Optional[bool] = True,
        save_scheduler: Optional[bool] = True,
        save_checkpoint: Optional[str] = "default",
        log_freq: Optional[int] = 100,
        ckpt_freq: Optional[int] = 10000,
        grad_acc_step: Optional[int] = 1,
        max_grad_norm: Optional[float] = 1.0,
        num_training_samples: Optional[int] = 1000000000,
        epochs: Optional[int] = 5,
        use_ema: Optional[bool] = False,
        ema_decay: Optional[float] = 0.9999,
        ema_tau: Optional[int] = 2000,
        use_amp: Optional[bool] = True,
    ):
        """
        Train the model.

        Args:
            optim: The optimizer for training.
            loss_fn: The loss function for training.
            score_fn: The scoring function for evaluation.
            monitor_fns (optional): The monitoring functions for evaluation. Defaults to None.
            scheduler (optional): The scheduler for adjusting the learning rate. Defaults to None.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            to_ckpt_dir (optional): The directory path to save checkpoints to. Defaults to "./to_ckpt".
            train_batch_size (optional): The batch size for training. Defaults to 128.
            dev_batch_size (optional): The batch size for evaluation. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            save_optimizer (optional): Whether to save the optimizer. Defaults to True.
            save_scheduler (optional): Whether to save the scheduler. Defaults to True.
            log_freq (optional): The frequency of logging training information. Defaults to 100.
            ckpt_freq (optional): The frequency of saving checkpoints. Defaults to 10000.
            grad_acc_step (optional): The number of gradient accumulation steps. Defaults to 1.
            max_grad_norm (optional): The maximum gradient norm for gradient clipping. Defaults to 1.0.
            num_training_samples (optional): The number of training samples. Defaults to 1000000000.
            epochs (optional): The number of training epochs. Defaults to 5.
            use_ema (optional): Whether to use exponential moving average. Defaults to False.
            ema_decay (optional): The decay rate for exponential moving average. Defaults to 0.9999.
            ema_tau (optional): The time constant for exponential moving average. Defaults to 2000.
            gpu_mode (optional): Whether to make GPU active. Defaults to False.
        """
        if not os.path.exists(to_ckpt_dir) and self.local_rank in [-1, 0]:
            os.makedirs(to_ckpt_dir, exist_ok=True)

        if loss_fn is not None:
            loss_fn = init_registered_module(loss_fn, self.config, registered_loss)

        if score_fn is not None:
            score_fn = init_registered_module(score_fn, self.config, registered_score)

        if monitor_fns is not None:
            monitor_fns = [
                init_registered_module(monitor_fn, self.config, registered_score)
                for monitor_fn in monitor_fns
                if monitor_fn in registered_score
            ]

        if optim is not None and self.model is not None:
            optim = init_registered_module(
                optim,
                self.config,
                registered_optim,
                params=filter(lambda x: x.requires_grad, self.model.parameters()),
            )

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)
            optim.from_checkpoint(
                from_ckpt_dir,
                weight_name="pytorch_optim.bin",
            )

        if os.path.exists(to_ckpt_dir):
            self.model.from_checkpoint(
                to_ckpt_dir,
                weight_name="pytorch_model_latest.bin",
            )
            optim.from_checkpoint(
                to_ckpt_dir,
                weight_name="pytorch_optim_latest.bin",
            )

        info_path = os.path.join(to_ckpt_dir, "info.json")
        if os.path.exists(info_path):
            info = json.load(open(os.path.join(to_ckpt_dir, "info.json")))
        else:
            info = dict()

        global_epoch = info.get("global_epoch", 0)
        global_step = info.get("global_step", 0)
        self.best_score = info.get("best_score", self.best_score)

        logging.info(f"the best score is {self.best_score}")

        self.ema_model = None
        if use_ema:
            num_ema_steps = info.get("num_ema_steps", 0)
            self.ema_model = ExponentialMovingAverage(
                self.model,
                decay=ema_decay,
                tau=ema_tau,
                num_steps=num_ema_steps,
            )
            if os.path.exists(from_ckpt_dir):
                self.ema_model.from_checkpoint(
                    from_ckpt_dir,
                    weight_name="pytorch_ema_model.bin",
                )
            if os.path.exists(to_ckpt_dir):
                self.ema_model.from_checkpoint(
                    to_ckpt_dir,
                    weight_name="pytorch_ema_model_latest.bin",
                )

        for n, p in self.model.named_parameters():
            logging.debug(
                f"{n}: trainable - {p.requires_grad} | tensor dtype - {p.dtype} | tensor shape - {p.shape} | tensor device - {p.device}"
            )

        global_rank = -1
        if self.n_gpu > 1:
            self.model = nn.parallel.DistributedDataParallel(
                self.model,
                device_ids=[self.local_rank],
                output_device=self.local_rank,
                find_unused_parameters=True,
                broadcast_buffers=False,
            )
            global_rank = dist.get_rank()

        train_sampler = DistributedSkipSampler if self.n_gpu > 1 else RandomSkipSampler
        dev_sampler = DistributedSampler if self.n_gpu > 1 else SequentialSampler

        dataset_train = self.datasets.get("train")
        dataset_dev = self.datasets.get("dev")

        iter_train = DataLoader(
            dataset_train,
            sampler=train_sampler(dataset_train)
            if not isinstance(dataset_train, Iterable)
            else None,
            batch_size=train_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        iter_dev = DataLoader(
            dataset_dev,
            sampler=dev_sampler(dataset_dev)
            if not isinstance(dataset_dev, Iterable)
            else None,
            batch_size=dev_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        if scheduler is not None:
            if not isinstance(dataset_train, Iterable):
                num_training_steps = int(
                    epochs
                    * len(dataset_train)
                    // train_batch_size
                    // max(1, self.n_gpu)
                    // grad_acc_step
                )
            else:
                num_training_steps = int(
                    epochs
                    * num_training_samples
                    // train_batch_size
                    // max(1, self.n_gpu)
                    // grad_acc_step
                )

            scheduler = init_registered_module(
                scheduler,
                self.config,
                registered_scheduler,
                optimizer=optim,
                num_training_steps=num_training_steps,
            )

        if scheduler and os.path.exists(to_ckpt_dir):
            scheduler.from_checkpoint(
                to_ckpt_dir,
                weight_name="pytorch_scheduler_latest.bin",
            )

        if use_amp:
            scaler = GradScaler()

        log_loss = 0
        dev_epoch = 0
        for e in range(0, epochs):
            torch.cuda.empty_cache()
            if e < global_epoch:
                continue

            if hasattr(dataset_train, "set_epoch"):
                dataset_train.set_epoch(e)

            if hasattr(dataset_train, "set_skip_step"):
                dataset_train.set_skip_step(global_step * train_batch_size)

            if hasattr(iter_train.sampler, "set_epoch"):
                iter_train.sampler.set_epoch(e)

            if hasattr(iter_train.sampler, "set_skip_step"):
                iter_train.sampler.set_skip_step(global_step * train_batch_size)

            self.model.train()
            is_update_step = True
            for step, (inputs, targets) in enumerate(iter_train):
                step = step + global_step
                is_update_step = False
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                    targets = targets.cuda()

                with torch.autocast(
                    enabled=use_amp,
                    device_type="cuda" if torch.cuda.is_available() else "cpu",
                ):
                    outputs = self.model(**inputs.dict())
                    if isinstance(outputs, LossOutputs):
                        loss = outputs.loss / grad_acc_step
                    else:
                        loss = loss_fn(outputs=outputs, targets=targets) / grad_acc_step

                nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                if use_amp:
                    scaler.scale(loss).backward()
                else:
                    loss.backward()
                log_loss += loss.data * grad_acc_step
                if (step + 1) % grad_acc_step == 0:
                    is_update_step = True
                    if use_amp:
                        scaler.step(optim)
                        scaler.update()
                    else:
                        optim.step()
                    if scheduler is not None:
                        scheduler.step()
                    optim.zero_grad()

                    if use_ema and self.ema_model is not None:
                        self.ema_model.step(
                            self.model.module if self.n_gpu > 1 else self.model
                        )

                if (step + 1) % log_freq == 0 and global_rank in [-1, 0]:
                    logging.info(
                        f"epoch {e} step {step}: loss -- { log_loss / log_freq }"
                    )
                    if wandb.is_available():
                        wandb.log(
                            {
                                "epoch": e,
                                "step": step,
                                "loss": log_loss / log_freq,
                            }
                        )
                    log_loss = 0

                if (step + 1) % ckpt_freq == 0:
                    if hasattr(dataset_dev, "set_epoch"):
                        dataset_dev.set_epoch(dev_epoch)

                    if hasattr(iter_dev.sampler, "set_epoch"):
                        iter_dev.sampler.set_epoch(dev_epoch)

                    dev_epoch += 1
                    self.best_score = save_snapshot(
                        self.model.module if self.n_gpu > 1 else self.model,
                        to_ckpt_dir,
                        iter_dev,
                        score_fn,
                        monitor_fns,
                        optim=optim if save_optimizer else None,
                        scheduler=scheduler if save_scheduler else None,
                        save_checkpoint=save_checkpoint,
                        ema_model=self.ema_model if use_ema else None,
                        best_score=self.best_score,
                        info_path=info_path,
                        local_rank=self.local_rank,
                        global_epoch=e,
                        global_step=step + 1,
                    )

            if not is_update_step:
                scaler.step(optim)
                scaler.update()
                if scheduler is not None:
                    scheduler.step()
                optim.zero_grad()

                if use_ema and self.ema_model is not None:
                    self.ema_model.step(
                        self.model.module if self.n_gpu > 1 else self.model
                    )

            log_loss = 0

            if hasattr(dataset_dev, "set_epoch"):
                dataset_dev.set_epoch(dev_epoch)

            if hasattr(iter_dev.sampler, "set_epoch"):
                iter_dev.sampler.set_epoch(dev_epoch)

            dev_epoch += 1

            global_step = 0
            self.best_score = save_snapshot(
                self.model.module if self.n_gpu > 1 else self.model,
                to_ckpt_dir,
                iter_dev,
                score_fn,
                monitor_fns,
                optim=optim if save_optimizer else None,
                scheduler=scheduler if save_scheduler else None,
                save_checkpoint=save_checkpoint,
                ema_model=self.ema_model if use_ema else None,
                best_score=self.best_score,
                info_path=info_path,
                local_rank=self.local_rank,
                global_epoch=e + 1,
                global_step=0,
            )

    @torch.no_grad()
    @add_default_section_for_function("core/task/supervised")
    def eval(
        self,
        monitor_fns: Union[str, List[str]],
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        dev_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
    ):
        """
        Perform evaluation on the model.

        Args:
            monitor_fns: The monitoring functions for evaluation.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            dev_batch_size (optional): The batch size for evaluation. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
        """
        monitor_fns = [
            init_registered_module(monitor_fn, self.config, registered_score)
            for monitor_fn in monitor_fns
            if monitor_fn in registered_score
        ]

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        global_rank = -1
        if self.n_gpu > 1:
            self.model = nn.parallel.DistributedDataParallel(
                self.model,
                device_ids=[self.local_rank],
                output_device=self.local_rank,
                find_unused_parameters=False,
                broadcast_buffers=False,
            )
            global_rank = dist.get_rank()

        dev_sampler = DistributedSampler if self.n_gpu > 1 else SequentialSampler
        dataset_dev = self.datasets.get("dev")
        iter_dev = DataLoader(
            dataset_dev,
            sampler=dev_sampler(dataset_dev)
            if not isinstance(dataset_dev, Iterable)
            else None,
            batch_size=dev_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        results = infer(self.model.module if self.n_gpu > 1 else self.model, iter_dev)
        if global_rank in [-1, 0]:
            monitor(
                outputs=results.outputs,
                targets=results.targets,
                monitor_fns=monitor_fns,
            )

    @torch.no_grad()
    @add_default_section_for_function("core/task/supervised")
    def infer(
        self,
        postprocess_fn: str,
        writer: str,
        test_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        max_size: Optional[int] = 10000,
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        output_header: Optional[List] = None,
        output_path: Optional[str] = "./output.txt",
        postprocess_workers: Optional[int] = 2,
    ):
        """
        Perform inference using the model.

        Args:
            postprocess_fn: The postprocessing function for inference.
            writer: The writer to save the inference results.
            test_batch_size (optional): The batch size for inference. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            max_size (optional): The maximum number of samples to process. Defaults to 10000.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            output_header (optional): The header for the output file. Defaults to None.
            output_path (optional): The path to save the output file. Defaults to "./output.txt".
            postprocess_workers (optional): The number of worker processes for postprocessing. Defaults to 2.
        """
        assert self.n_gpu <= 1
        assert writer is not None

        output_dir = os.path.dirname(output_path)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        if postprocess_fn is not None:
            postprocess_fn = init_registered_process(postprocess_fn, self.config)

        if writer is not None:
            writer = init_registered_module(
                writer,
                self.config,
                registered_writer,
                output_file=output_path,
            )

        skip_step = writer.skip_n_samples

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        if skip_step == 0:
            sampler = SequentialSampler
        else:
            sampler = SequentialSkipSampler

        dataset_test = self.datasets.get("test")

        iter_test = DataLoader(
            dataset_test,
            sampler=sampler(dataset_test)
            if not isinstance(dataset_test, Iterable)
            else None,
            batch_size=test_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        if skip_step > 0 and hasattr(dataset_test, "set_skip_step"):
            dataset_test.set_skip_step(skip_step)

        if skip_step > 0 and hasattr(iter_test.sampler, "set_skip_step"):
            iter_test.sampler.set_skip_step(skip_step)

        if hasattr(dataset_test, "dataset"):
            data_info = dataset_test.dataset
            data_info = DatasetFeature(data_info)
            iter_data = DataLoader(
                deepcopy(data_info),
                sampler=sampler(data_info)
                if not isinstance(dataset_test, Iterable)
                else None,
                batch_size=test_batch_size,
                shuffle=False,
                pin_memory=pin_memory,
                num_workers=num_workers,
                collate_fn=None,
            )
        else:
            iter_data = None

        if skip_step > 0 and hasattr(iter_data.sampler, "set_skip_step"):
            iter_data.sampler.set_skip_step(skip_step)

        self.model.eval()
        start = time.time()

        data_queue = Queue(maxsize=max_size)
        msg_queue = Queue(maxsize=max_size)
        postprocess_list = []
        for _ in range(postprocess_workers):
            p = PostProcess(
                postprocess_fn,
                data_queue,
                msg_queue,
            )
            postprocess_list.append(p)
            p.start()

        io_process = IOProcess(msg_queue, writer=writer)
        io_process.start()

        if iter_data is None:
            for step, (inputs, _) in enumerate(iter_test):
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                outputs = self.model(**inputs.dict())
                outputs = outputs.cpu()
                data_queue.put((step, outputs))
        else:
            for step, ((inputs, _), _infos) in enumerate(zip(iter_test, iter_data)):
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                outputs = self.model(**inputs.dict())
                outputs = outputs.cpu()
                if output_header is not None:
                    _infos = {k: _infos[k] for k in output_header if k in _infos}
                    outputs.from_pandas(pd.DataFrame(_infos))
                data_queue.put((step, outputs))

        data_queue.put((-1, GENERATE_FINISHED))
        for p in postprocess_list:
            p.join()

        msg_queue.put((-1, GENERATE_FINISHED))
        io_process.join()

        end = time.time()
        ms = (end - start) * 1000
        logging.info(
            "{:.2f} ms, {:.2f} sample/s".format(
                ms,
                ((len(dataset_test) - skip_step) / ms * 1000),
            )
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import re
import gc
import json
import logging
import torch
import hashlib
import asyncio
import pandas as pd
from PIL import Image
from fastapi import APIRouter, UploadFile, File
from fastapi.responses import StreamingResponse
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil
from diffusers.models import ControlNetModel
from diffusers.pipelines import (
    StableDiffusionPipeline,
    StableDiffusionImg2ImgPipeline,
    StableDiffusionInpaintPipeline,
    StableDiffusionUpscalePipeline,
    StableDiffusionDepth2ImgPipeline,
    StableVideoDiffusionPipeline,
    StableDiffusionControlNetPipeline,
    StableDiffusionControlNetImg2ImgPipeline,
    StableDiffusionControlNetInpaintPipeline,
)
from unitorch import is_xformers_available
from unitorch.utils import is_remote_url
from unitorch.models.diffusers import GenericStableModel
from unitorch.models.diffusers import StableProcessor

from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    register_fastapi,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericFastAPI
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
    load_weight,
)
from unitorch.cli.pipelines import Schedulers
from unitorch.cli.pipelines.stable.interrogator import ClipInterrogatorPipeline


@register_fastapi("core/fastapi/interrogator/clip")
class ClipInterrogatorFastAPI(GenericFastAPI):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section(f"core/fastapi/interrogator/clip")
        router = config.getoption("router", "/core/fastapi/interrogator/clip")
        self._pipe = None
        self._router = APIRouter(prefix=router)
        self._router.add_api_route("/generate", self.serve, methods=["POST"])
        self._router.add_api_route("/status", self.status, methods=["GET"])
        self._router.add_api_route("/start", self.start, methods=["GET"])
        self._router.add_api_route("/stop", self.stop, methods=["GET"])
        self._lock = asyncio.Lock()

    @property
    def router(self):
        return self._router

    def start(self):
        self._pipe = ClipInterrogatorPipeline.from_core_configure(self.config)
        return "start success"

    def stop(self):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        return "stop success"

    def status(self):
        return "running" if self._pipe is not None else "stopped"

    async def serve(
        self,
        image: UploadFile,
    ):
        assert self._pipe is not None
        image_bytes = await image.read()
        image = Image.open(io.BytesIO(image_bytes))
        async with self._lock:
            outputs = self._pipe(image)

        return outputs.best_prompt



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_clip_infos = {
    "clip-vit-base-patch16": {
        "config": hf_endpoint_url(
            "/openai/clip-vit-base-patch16/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/openai/clip-vit-base-patch16/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/openai/clip-vit-base-patch16/resolve/main/vocab.json"
        ),
        "merge": hf_endpoint_url(
            "/openai/clip-vit-base-patch16/resolve/main/merges.txt"
        ),
        "weight": hf_endpoint_url(
            "/openai/clip-vit-base-patch16/resolve/main/pytorch_model.bin"
        ),
    },
    "clip-vit-base-patch32": {
        "config": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/vocab.json"
        ),
        "merge": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/merges.txt"
        ),
        "weight": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/pytorch_model.bin"
        ),
    },
    "clip-vit-large-patch14": {
        "config": hf_endpoint_url(
            "/openai/clip-vit-large-patch14/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/openai/clip-vit-large-patch14/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/openai/clip-vit-large-patch14/resolve/main/vocab.json"
        ),
        "merge": hf_endpoint_url(
            "/openai/clip-vit-large-patch14/resolve/main/merges.txt"
        ),
        "weight": hf_endpoint_url(
            "/openai/clip-vit-large-patch14/resolve/main/pytorch_model.bin"
        ),
    },
    # https://gist.github.com/rwightman/c79fd0241ed3c860e898114931c07990
    "clip-vit-huge-patch14-hpsv2": {
        "config": hf_endpoint_url("/adams-story/HPSv2-hf/resolve/main/config.json"),
        "vision_config": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/vocab.json"
        ),
        "merge": hf_endpoint_url(
            "/openai/clip-vit-base-patch32/resolve/main/merges.txt"
        ),
        "weight": hf_endpoint_url(
            "/adams-story/HPSv2-hf/resolve/main/pytorch_model.bin"
        ),
    },
}

pretrained_clip_extensions_infos = {}

import unitorch.cli.models.clip.modeling
import unitorch.cli.models.clip.processing
from unitorch.cli.models.clip.modeling import (
    ClipForPretrain,
    ClipForClassification,
    ClipForTextClassification,
    ClipForImageClassification,
)
from unitorch.cli.models.clip.processing import ClipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models.detr import (
    DetrForDetection as _DetrForDetection,
)
from unitorch.models.detr import DetrProcessor
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericScript
from unitorch.cli import register_script
from unitorch.cli.models.detr import pretrained_detr_infos


class DetrForDetectionPipeline(_DetrForDetection):
    def __init__(
        self,
        config_path: str,
        vision_config_path: str,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        enable_cpu_offload: Optional[bool] = True,
        device: Optional[Union[str, int]] = "cpu",
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = DetrProcessor(
            vision_config_path=vision_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self._enable_cpu_offload = enable_cpu_offload
        if not self._enable_cpu_offload and self._device != "cpu":
            self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/detr")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        config.set_default_section("core/pipeline/detr")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "detr-resnet-50"
        )

        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vision_config_path = vision_config_path or config.getoption(
            "vision_config_path", None
        )
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        device = config.getoption("device", "cpu") if device is None else device
        pretrained_weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vision_config_path,
            weight_path=weight_path,
            enable_cpu_offload=enable_cpu_offload,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/detr")
    def __call__(
        self,
        image: Union[Image.Image, str],
        threshold: Optional[float] = 0.5,
    ):
        if self._enable_cpu_offload:
            self.to(self._device)
        inputs = self.processor.image(image)
        pixel_values = [inputs.image.to(self._device)]
        outputs = self.detect(
            pixel_values,
            norm_bboxes=True,
            threshold=threshold,
        )
        result_image = image.copy()
        bboxes = outputs["bboxes"][0].cpu().numpy()
        scores = outputs["scores"][0].cpu().numpy()
        classes = outputs["classes"][0].cpu().numpy()
        # draw boxes on the result image with class and score
        for bbox, score, classid in zip(bboxes, scores, classes):
            if score < threshold:
                continue
            bbox = bbox * np.array(
                [
                    result_image.width,
                    result_image.height,
                    result_image.width,
                    result_image.height,
                ]
            )
            bbox = list(map(int, bbox))
            result_image = result_image.copy()
            draw = ImageDraw.Draw(result_image)
            draw.rectangle(bbox, outline="red", width=3)
            draw.text((bbox[0] + 5, bbox[1] + 5), f"{classid}", fill="blue")
            draw.text((bbox[0] + 5, bbox[1] + 15), f"{score:.2f}", fill="green")

        if self._enable_cpu_offload:
            self.to("cpu")
            torch.cuda.empty_cache()
        return result_image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import io
import requests
import time
import base64
import json
import logging
import torch
import torch.nn as nn
import numpy as np
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from random import random

ACT2FN = {
    "relu": lambda v: np.maximum(v, 0),
    "tanh": np.tanh,
    "sigmoid": lambda v: 1 / (1 + np.exp(-v)),
    "softmax": lambda v: np.exp(v) / np.sum(np.exp(v), axis=-1, keepdims=True),
}
from unitorch import is_diffusers_available, is_opencv_available, is_megatron_available

# import modeling utils
from unitorch.cli.models.modeling_utils import (
    ModelInputs,
    ModelOutputs,
    ModelTargets,
    TensorsInputs,
    ListTensorsInputs,
    CombineTensorsInputs,
    TensorsOutputs,
    ListTensorsOutputs,
    CombineTensorsOutputs,
    TensorsTargets,
    ListTensorsTargets,
    CombineTensorsTargets,
    LossOutputs,
)
from unitorch.cli.models.classification_utils import (
    ClassificationOutputs,
    EmbeddingOutputs,
    ClassificationTargets,
)
from unitorch.cli.models.detection_utils import (
    DetectionOutputs,
    DetectionTargets,
)
from unitorch.cli.models.detection_utils import detection_model_decorator
from unitorch.cli.models.generation_utils import (
    GenerationOutputs,
    GenerationTargets,
)
from unitorch.cli.models.generation_utils import generation_model_decorator
from unitorch.cli.models.ranking_utils import RankingOutputs, RankingTargets
from unitorch.cli.models.segmentation_utils import (
    SegmentationOutputs,
    SegmentationTargets,
)
from unitorch.cli.models.segmentation_utils import segmentation_model_decorator

# import processing utils
import unitorch.cli.models.image_utils
import unitorch.cli.models.random_utils
import unitorch.cli.models.label_utils
import unitorch.cli.models.processing_utils

if is_opencv_available():
    import unitorch.cli.models.video_utils

if is_diffusers_available():
    from unitorch.cli.models.diffusion_utils import DiffusionOutputs, DiffusionTargets
    from unitorch.cli.models.diffusion_utils import diffusion_model_decorator
    import unitorch.cli.models.diffusers

if is_megatron_available():
    import unitorch.cli.models.megatron

# import model classes & process functions
import unitorch.cli.models.bart
import unitorch.cli.models.bert
import unitorch.cli.models.beit
import unitorch.cli.models.blip
import unitorch.cli.models.bloom
import unitorch.cli.models.bria
import unitorch.cli.models.chinese_clip
import unitorch.cli.models.clip
import unitorch.cli.models.detr
import unitorch.cli.models.dinov2
import unitorch.cli.models.dpt
import unitorch.cli.models.grounding_dino
import unitorch.cli.models.kolors
import unitorch.cli.models.llama
import unitorch.cli.models.llava
import unitorch.cli.models.mask2former
import unitorch.cli.models.mbart
import unitorch.cli.models.mistral
import unitorch.cli.models.mt5
import unitorch.cli.models.pegasus
import unitorch.cli.models.peft
import unitorch.cli.models.roberta
import unitorch.cli.models.swin
import unitorch.cli.models.sam
import unitorch.cli.models.siglip
import unitorch.cli.models.segformer
import unitorch.cli.models.t5
import unitorch.cli.models.visualbert
import unitorch.cli.models.vit
import unitorch.cli.models.xlm_roberta
import unitorch.cli.models.xpegasus



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import re
import gc
import json
import logging
import torch
import hashlib
import asyncio
import pandas as pd
from PIL import Image
from fastapi import APIRouter, UploadFile, File
from fastapi.responses import StreamingResponse
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch import is_xformers_available
from unitorch.utils import is_remote_url
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    register_fastapi,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericFastAPI
from unitorch.cli.pipelines.llava import (
    LlavaMistralClipForGenerationPipeline,
    LlavaLlamaSiglipForGenerationPipeline,
)


@register_fastapi("core/fastapi/llava/mistral_clip")
class LlavaMistralClipFastAPI(GenericFastAPI):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section(f"core/fastapi/llava/mistral_clip")
        router = config.getoption("router", "/core/fastapi/llava/mistral_clip")
        self._pipe = None
        self._router = APIRouter(prefix=router)
        self._router.add_api_route("/generate", self.serve, methods=["POST"])
        self._router.add_api_route("/status", self.status, methods=["GET"])
        self._router.add_api_route("/start", self.start, methods=["GET"])
        self._router.add_api_route("/stop", self.stop, methods=["GET"])
        self._lock = asyncio.Lock()

    @property
    def router(self):
        return self._router

    def start(self):
        self._pipe = LlavaMistralClipForGenerationPipeline.from_core_configure(
            self.config,
            pretrained_name="llava-v1.6-mistral-7b-hf",
        )
        return "start success"

    def stop(self):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        return "stop success"

    def status(self):
        return "running" if self._pipe is not None else "stopped"

    async def serve(
        self,
        text: str,
        image: UploadFile,
    ):
        assert self._pipe is not None
        image_bytes = await image.read()
        image = Image.open(io.BytesIO(image_bytes))
        text = f"[INST] <image>\n {text} [/INST]"
        async with self._lock:
            caption = self._pipe(
                text,
                image,
                lora_checkpoints=[],
                lora_weights=[],
                lora_alphas=[],
                lora_urls=[],
                lora_files=[],
            )

        return caption


@register_fastapi("core/fastapi/llava/joycaption2")
class LlavaLlamaSiglipFastAPI(GenericFastAPI):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section(f"core/fastapi/llava/joycaption2")
        router = config.getoption("router", "/core/fastapi/llava/joycaption2")
        self._pipe = None
        self._router = APIRouter(prefix=router)
        self._router.add_api_route("/generate", self.serve, methods=["POST"])
        self._router.add_api_route("/status", self.status, methods=["GET"])
        self._router.add_api_route("/start", self.start, methods=["GET"])
        self._router.add_api_route("/stop", self.stop, methods=["GET"])
        self._lock = asyncio.Lock()

    @property
    def router(self):
        return self._router

    def start(self):
        self._pipe = LlavaLlamaSiglipForGenerationPipeline.from_core_configure(
            self.config,
            pretrained_name="llava-v1.6-joycaption-2",
        )
        return "start success"

    def stop(self):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        return "stop success"

    def status(self):
        return "running" if self._pipe is not None else "stopped"

    async def serve(
        self,
        text: str,
        image: UploadFile,
    ):
        assert self._pipe is not None
        image_bytes = await image.read()
        image = Image.open(io.BytesIO(image_bytes))
        text = f"<|start_header_id|>system<|end_header_id|>\\n\\nCutting Knowledge Date: December 2023\\nToday Date: 26 July 2024\\n\\nYou are a helpful image captioner.<|eot_id|><|start_header_id|>user<|end_header_id|>\\n\\n<|reserved_special_token_70|><|reserved_special_token_69|><|reserved_special_token_71|>{text}|eot_id|><|start_header_id|>assistant<|end_header_id|>\\n\\n"
        async with self._lock:
            caption = self._pipe(
                text,
                image,
                lora_checkpoints=[],
                lora_weights=[],
                lora_alphas=[],
                lora_urls=[],
                lora_files=[],
            )

        return caption



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import pandas as pd
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models import GenericOutputs
from unitorch.models.clip import (
    ClipForPretrain as _ClipForPretrain,
    ClipProcessor,
)
from unitorch.utils import pop_value, nested_dict_value, read_file, read_json_file

from unitorch.cli import (
    cached_path,
    hf_endpoint_url,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericScript
from unitorch.cli import register_script
from unitorch.cli.models.clip import pretrained_clip_infos
from unitorch.cli.pipelines.blip import BlipForImageCaptionPipeline


class ClipInterrogatorPipeline(_ClipForPretrain):
    def __init__(
        self,
        config_path: str,
        vocab_path: str,
        merge_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 77,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
    ):
        projection_dim = nested_dict_value(
            read_json_file(config_path), "projection_dim"
        )
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
        )
        self.processor = ClipProcessor(
            vocab_path=vocab_path,
            merge_path=merge_path,
            vision_config_path=vision_config_path,
            max_seq_length=max_seq_length,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.to(device=self._device)

        config = CoreConfigureParser()
        self.blip_pipe = BlipForImageCaptionPipeline.from_core_configure(
            config, pretrained_name="blip-image-captioning-large"
        )
        self.blip_pipe._device = self._device
        self.blip_pipe.to(device=self._device)

        artists = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/artists.txt"
                )
            ),
            lines=True,
        )
        flavors = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/flavors.txt"
                )
            ),
            lines=True,
        )
        mediums = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/mediums.txt"
                )
            ),
            lines=True,
        )
        movements = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/movements.txt"
                )
            ),
            lines=True,
        )
        negative = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/negative.txt"
                )
            ),
            lines=True,
        )
        sites = read_file(
            cached_path(
                hf_endpoint_url(
                    "/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/sites.txt"
                )
            ),
            lines=True,
        )

        artists = [f"by {a}" for a in artists] + [f"inspired by {a}" for a in artists]
        trendings = (
            sites
            + [f"trending on {s}" for s in sites]
            + [f"featured on {s}" for s in sites]
            + [f"{s} contest winner" for s in sites]
        )

        self.positive_labels = artists + flavors + mediums + movements + trendings
        self.positive_artists_labels = artists
        self.positive_flavors_labels = flavors
        self.positive_mediums_labels = mediums
        self.positive_movements_labels = movements
        self.positive_trendings_labels = trendings
        self.negative_labels = negative

        self.eval()

        self.positive_labels_embeds = self.get_text_embeds(self.positive_labels)

    @classmethod
    @add_default_section_for_init("core/pipeline/interrogator/clip")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        merge_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        config.set_default_section("core/pipeline/interrogator/clip")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "clip-vit-base-patch16"
        )

        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vocab_path = vocab_path or config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        merge_path = merge_path or config.getoption("merge_path", None)
        merge_path = pop_value(
            merge_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "merge"),
        )
        merge_path = cached_path(merge_path)

        vision_config_path = vision_config_path or config.getoption(
            "vision_config_path", None
        )
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        max_seq_length = config.getoption("max_seq_length", 77)
        device = config.getoption("device", "cpu") if device is None else device
        pretrained_weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vocab_path,
            merge_path,
            vision_config_path,
            max_seq_length=max_seq_length,
            weight_path=weight_path,
            device=device,
        )

        return inst

    def get_image_embeds(self, image: Image.Image):
        inputs = self.processor.image_classification(image)
        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self._device) if v is not None else v
            for k, v in inputs.items()
        }
        vision_outputs = self.vision_model(pixel_values=inputs["pixel_values"])
        image_embeds = self.visual_projection(vision_outputs[1])
        image_embeds = image_embeds / image_embeds.norm(dim=-1, keepdim=True)
        return image_embeds.cpu()

    def get_text_embeds(self, texts: Union[str, List[str]], max_batch_size=1024):
        if isinstance(texts, str):
            texts = [texts]
        inputs = [self.processor.text_classification(text) for text in texts]
        keys = inputs[0].keys()
        inputs = {
            k: torch.stack([i[k] for i in inputs]).to(device=self._device) for k in keys
        }
        results = []
        for i in range(0, len(inputs["input_ids"]), max_batch_size):
            text_outputs = self.text_model(
                input_ids=inputs["input_ids"][i : i + max_batch_size],
                attention_mask=inputs["attention_mask"][i : i + max_batch_size],
                position_ids=inputs["position_ids"][i : i + max_batch_size],
            )
            text_embeds = self.text_projection(text_outputs[1])
            text_embeds = text_embeds / text_embeds.norm(dim=-1, keepdim=True)
            results.append(text_embeds.cpu())
        return torch.cat(results, dim=0)

    def get_score(self, image_embeds, text):
        text_embeds = self.get_text_embeds(text)
        scores = torch.einsum("te,be->tb", text_embeds, image_embeds)
        return scores[0][0].item()

    def rank_top(self, image_embeds, texts=[], topk=1, reverse=True, text_embeds=None):
        if text_embeds is None:
            text_embeds = self.get_text_embeds(texts)
        scores = torch.einsum("te,be->tb", text_embeds, image_embeds)
        scores = scores[:, 0].tolist()
        return sorted(list(zip(texts, scores)), key=lambda x: x[1], reverse=reverse)[
            :topk
        ]

    def chain(
        self,
        image_embeds,
        phrases,
        caption=None,
        min_count=8,
        max_count=32,
        reverse=True,
    ):
        if caption is None:
            caption = self.rank_top(image_embeds, phrases, topk=1, reverse=reverse)[0][
                0
            ]
        best_prompt, best_score = caption, self.get_score(image_embeds, caption)
        if best_prompt in phrases:
            phrases.remove(best_prompt)

        prompt, score = best_prompt, best_score

        for i in range(max_count):
            new_prompt, new_score = self.rank_top(
                image_embeds,
                [f"{prompt}, {p}" for p in phrases],
                topk=1,
                reverse=reverse,
            )[0]

            label = new_prompt[len(prompt) + 2 :]
            phrases.remove(label)

            if (reverse and new_score > best_score) or (
                not reverse and new_score < best_score
            ):
                best_prompt, best_score = new_prompt, new_score
            if (
                i < min_count
                or (reverse and new_score < score)
                or (not reverse and new_score > score)
            ):
                prompt, score = new_prompt, new_score
            else:
                break

        return best_prompt

    def get_best_prompt(self, image_embeds, caption="", min_count=8, max_count=32):
        positive_labels = self.rank_top(
            image_embeds,
            texts=self.positive_labels,
            topk=1024,
            text_embeds=self.positive_labels_embeds,
        )
        positive_labels = list(map(lambda x: x[0], positive_labels))

        best_prompt = self.chain(
            image_embeds, positive_labels, caption, min_count, max_count
        )
        return best_prompt

    def get_fast_prompt(self, image_embeds, caption="", max_count=32):
        positive_labels = self.rank_top(
            image_embeds,
            texts=self.positive_labels,
            topk=max_count,
            text_embeds=self.positive_labels_embeds,
        )
        positive_labels = list(map(lambda x: x[0], positive_labels))

        fast_prompt = f"{caption}, {', '.join(positive_labels)}"
        return fast_prompt

    def get_classic_prompt(self, image_embeds, caption="", max_count=3):
        positive_embeds = self.positive_labels_embeds.clone()
        positive_artists_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_artists_labels)],
            positive_embeds[len(self.positive_artists_labels) :],
        )
        positive_flavors_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_flavors_labels)],
            positive_embeds[len(self.positive_flavors_labels) :],
        )
        positive_mediums_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_mediums_labels)],
            positive_embeds[len(self.positive_mediums_labels) :],
        )
        positive_movements_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_movements_labels)],
            positive_embeds[len(self.positive_movements_labels) :],
        )
        positive_trendings_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_trendings_labels)],
            positive_embeds[len(self.positive_trendings_labels) :],
        )

        artist = self.rank_top(
            image_embeds,
            texts=self.positive_artists_labels,
            topk=1,
            text_embeds=positive_artists_labels_embeds,
        )[0][0]
        flavors = self.rank_top(
            image_embeds,
            texts=self.positive_flavors_labels,
            topk=max_count,
            text_embeds=positive_flavors_labels_embeds,
        )
        flavors = ", ".join(list(map(lambda x: x[0], flavors)))
        medium = self.rank_top(
            image_embeds,
            texts=self.positive_mediums_labels,
            topk=1,
            text_embeds=positive_mediums_labels_embeds,
        )[0][0]
        movement = self.rank_top(
            image_embeds,
            texts=self.positive_movements_labels,
            topk=1,
            text_embeds=positive_movements_labels_embeds,
        )[0][0]
        trending = self.rank_top(
            image_embeds,
            texts=self.positive_trendings_labels,
            topk=1,
            text_embeds=positive_trendings_labels_embeds,
        )[0][0]

        if caption.startswith(medium):
            classic_prompt = f"{caption} {artist}, {trending}, {movement}, {flavors}"
        else:
            classic_prompt = (
                f"{caption}, {medium} {artist}, {trending}, {movement}, {flavors}"
            )

        return classic_prompt

    def get_negative_prompt(self, image_embeds, max_count=3):
        positive_flavors_labels_embeds = self.positive_labels_embeds[
            len(self.positive_artists_labels) : len(self.positive_artists_labels)
            + len(self.positive_flavors_labels)
        ]
        negative_labels = self.rank_top(
            image_embeds,
            texts=self.positive_flavors_labels,
            topk=max_count,
            reverse=False,
            text_embeds=positive_flavors_labels_embeds,
        )
        negative_labels = list(map(lambda x: x[0], negative_labels))
        negative_labels += self.negative_labels
        negative_prompt = self.chain(
            image_embeds, negative_labels, max_count=max_count, reverse=False
        )
        return negative_prompt

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/interrogator/clip")
    def __call__(
        self,
        image: Image.Image,
    ):
        caption = self.blip_pipe(image, num_beams=5, max_gen_seq_length=32)
        image_embeds = self.get_image_embeds(image)
        fast_prompt = self.get_fast_prompt(image_embeds, caption)
        classic_prompt = self.get_classic_prompt(image_embeds, caption)
        best_prompt = self.get_best_prompt(image_embeds, caption)
        negative_prompt = self.get_negative_prompt(image_embeds)
        return GenericOutputs(
            fast_prompt=fast_prompt,
            classic_prompt=classic_prompt,
            best_prompt=best_prompt,
            negative_prompt=negative_prompt,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import numpy as np
from PIL import Image, ImageFilter
from typing import Any, Dict, List, Optional, Tuple, Union
from unitorch.cli import CoreConfigureParser
from unitorch.cli.pipelines.dpt import DPTForDepthEstimationPipeline

dpt_pipe = None


def depth(image: Image.Image):
    global dpt_pipe
    if dpt_pipe is None:
        dpt_pipe = DPTForDepthEstimationPipeline.from_core_configure(
            CoreConfigureParser(), pretrained_name="dpt-large"
        )
        dpt_pipe.to("cpu")
    return dpt_pipe(image)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.utils import is_opencv_available
from unitorch.models.diffusers.modeling_stable import compute_snr
from unitorch.models.diffusers.modeling_stable import (
    GenericStableModel,
    StableForText2ImageGeneration,
    StableForImage2ImageGeneration,
    StableForImageInpainting,
    StableForImageResolution,
    StableForImage2VideoGeneration,
)
from unitorch.models.diffusers.modeling_stable_xl import (
    GenericStableXLModel,
    StableXLForText2ImageGeneration,
    StableXLForImage2ImageGeneration,
    StableXLForImageInpainting,
)

from unitorch.models.diffusers.modeling_stable_3 import (
    GenericStable3Model,
    Stable3ForText2ImageGeneration,
    Stable3ForImage2ImageGeneration,
    Stable3ForImageInpainting,
)
from unitorch.models.diffusers.modeling_stable_flux import (
    GenericStableFluxModel,
    StableFluxForText2ImageGeneration,
    StableFluxForImage2ImageGeneration,
    StableFluxForImageControlGeneration,
    StableFluxForImageReduxGeneration,
    StableFluxForImageInpainting,
)
from unitorch.models.diffusers.modeling_controlnet import (
    ControlNetForText2ImageGeneration,
    ControlNetForImage2ImageGeneration,
    ControlNetForImageInpainting,
)
from unitorch.models.diffusers.modeling_controlnet_xl import (
    ControlNetXLForText2ImageGeneration,
    ControlNetXLForImage2ImageGeneration,
    ControlNetXLForImageInpainting,
)

from unitorch.models.diffusers.modeling_controlnet_3 import (
    ControlNet3ForText2ImageGeneration,
    ControlNet3ForImageInpainting,
)
from unitorch.models.diffusers.modeling_controlnet_flux import (
    ControlNetFluxForText2ImageGeneration,
    ControlNetFluxForImage2ImageGeneration,
    ControlNetFluxForImageInpainting,
)
from unitorch.models.diffusers.modeling_adapter import (
    StableAdapterForText2ImageGeneration,
)
from unitorch.models.diffusers.modeling_adapter_xl import (
    StableXLAdapterForText2ImageGeneration,
)
from unitorch.models.diffusers.modeling_wan import (
    WanForText2VideoGeneration,
    WanForImage2VideoGeneration,
)
from unitorch.models.diffusers.modeling_vae import VAEForDiffusion
from unitorch.models.diffusers.processing_stable import (
    StableProcessor,
    StableVideoProcessor,
)
from unitorch.models.diffusers.processing_stable_xl import StableXLProcessor
from unitorch.models.diffusers.processing_stable_3 import Stable3Processor
from unitorch.models.diffusers.processing_stable_flux import StableFluxProcessor

if is_opencv_available():
    from unitorch.models.diffusers.processing_wan import WanProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import diffusers
from diffusers.pipelines import (
    StableDiffusion3Img2ImgPipeline,
    StableDiffusion3InpaintPipeline,
)
from diffusers.pipelines.stable_diffusion_3.pipeline_stable_diffusion_3_img2img import (
    SD3Transformer2DModel,
    FlowMatchEulerDiscreteScheduler,
    AutoencoderKL,
    CLIPTextModelWithProjection,
    CLIPTokenizer,
    T5EncoderModel,
    T5TokenizerFast,
    VaeImageProcessor,
)
from diffusers.pipelines.wan.pipeline_wan_i2v import (
    AutoTokenizer,
    CLIPImageProcessor,
    AutoencoderKLWan,
    UMT5EncoderModel,
    WanTransformer3DModel,
    CLIPVisionModel,
    WanImageToVideoPipeline,
)
from unitorch.models import GenericOutputs
from unitorch.utils.decorators import replace


@replace(
    diffusers.pipelines.stable_diffusion_3.pipeline_stable_diffusion_3_img2img.StableDiffusion3Img2ImgPipeline
)
class StableDiffusion3Img2ImgPipelineV2(StableDiffusion3Img2ImgPipeline):
    def __init__(
        self,
        transformer: SD3Transformer2DModel,
        scheduler: FlowMatchEulerDiscreteScheduler,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModelWithProjection,
        tokenizer: CLIPTokenizer,
        text_encoder_2: CLIPTextModelWithProjection,
        tokenizer_2: CLIPTokenizer,
        text_encoder_3: T5EncoderModel,
        tokenizer_3: T5TokenizerFast,
    ):
        super().__init__(
            transformer=transformer,
            scheduler=scheduler,
            vae=vae,
            text_encoder=text_encoder,
            tokenizer=tokenizer
            if tokenizer is not None
            else GenericOutputs(model_max_length=77),
            text_encoder_2=text_encoder_2,
            tokenizer_2=tokenizer_2
            if tokenizer_2 is not None
            else GenericOutputs(model_max_length=77),
            text_encoder_3=text_encoder_3,
            tokenizer_3=tokenizer_3
            if tokenizer_3 is not None
            else GenericOutputs(model_max_length=77),
        )


@replace(
    diffusers.pipelines.stable_diffusion_3.pipeline_stable_diffusion_3_inpaint.StableDiffusion3InpaintPipeline
)
class StableDiffusion3InpaintPipelineV2(StableDiffusion3InpaintPipeline):
    def __init__(
        self,
        transformer: SD3Transformer2DModel,
        scheduler: FlowMatchEulerDiscreteScheduler,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModelWithProjection,
        tokenizer: CLIPTokenizer,
        text_encoder_2: CLIPTextModelWithProjection,
        tokenizer_2: CLIPTokenizer,
        text_encoder_3: T5EncoderModel,
        tokenizer_3: T5TokenizerFast,
    ):
        super().__init__(
            transformer=transformer,
            scheduler=scheduler,
            vae=vae,
            text_encoder=text_encoder,
            tokenizer=tokenizer
            if tokenizer is not None
            else GenericOutputs(model_max_length=77),
            text_encoder_2=text_encoder_2,
            tokenizer_2=tokenizer_2
            if tokenizer_2 is not None
            else GenericOutputs(model_max_length=77),
            text_encoder_3=text_encoder_3,
            tokenizer_3=tokenizer_3
            if tokenizer_3 is not None
            else GenericOutputs(model_max_length=77),
        )


@replace(diffusers.pipelines.wan.pipeline_wan_i2v.WanImageToVideoPipeline)
class WanImageToVideoPipelineV2(WanImageToVideoPipeline):
    def __init__(
        self,
        tokenizer: AutoTokenizer,
        text_encoder: UMT5EncoderModel,
        image_encoder: CLIPVisionModel,
        image_processor: CLIPImageProcessor,
        transformer: WanTransformer3DModel,
        vae: AutoencoderKLWan,
        scheduler: FlowMatchEulerDiscreteScheduler,
    ):
        super().__init__(
            tokenizer=tokenizer,
            text_encoder=text_encoder,
            image_encoder=image_encoder,
            image_processor=image_processor,
            transformer=transformer,
            vae=vae,
            scheduler=scheduler,
        )

    def check_inputs(
        self,
        prompt,
        negative_prompt,
        image,
        height,
        width,
        prompt_embeds=None,
        negative_prompt_embeds=None,
        image_embeds=None,
        callback_on_step_end_tensor_inputs=None,
    ):
        pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.webuis import SimpleWebUI
from unitorch.cli.webuis.stable_flux.text2image import StableFluxText2ImageWebUI
from unitorch.cli.webuis.stable_flux.image2image import StableFluxImage2ImageWebUI
from unitorch.cli.webuis.stable_flux.image_control import StableFluxImageControlWebUI
from unitorch.cli.webuis.stable_flux.image_redux import StableFluxImageReduxWebUI
from unitorch.cli.webuis.stable_flux.inpainting import StableFluxImageInpaintingWebUI
from unitorch.cli.webuis.stable_flux.redux_inpainting import (
    StableFluxReduxInpaintingWebUI,
)


@register_webui("core/webui/stable_flux")
class StableFluxWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        webuis = [
            StableFluxText2ImageWebUI(config),
            StableFluxImage2ImageWebUI(config),
            StableFluxImageControlWebUI(config),
            StableFluxImageReduxWebUI(config),
            StableFluxImageInpaintingWebUI(config),
            StableFluxReduxInpaintingWebUI(config),
        ]
        iface = gr.TabbedInterface(
            [webui.iface for webui in webuis],
            tab_names=[webui.iname for webui in webuis],
        )
        super().__init__(config, iname="StableFlux", iface=iface)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_visualbert_infos = {
    "visualbert-vqa-coco-pre": {
        "config": hf_endpoint_url(
            "/uclanlp/visualbert-vqa-coco-pre/resolve/main/config.json"
        ),
        "vocab": hf_endpoint_url("/bert-base-uncased/resolve/main/vocab.txt"),
        "weight": hf_endpoint_url(
            "/uclanlp/visualbert-vqa-coco-pre/resolve/main/pytorch_model.bin"
        ),
    },
    "visualbert-nlvr2-coco-pre": {
        "config": hf_endpoint_url(
            "/uclanlp/visualbert-nlvr2-coco-pre/resolve/main/config.json"
        ),
        "vocab": hf_endpoint_url(
            "/bert-base-multilingual-uncased/resolve/main/vocab.txt"
        ),
        "weight": hf_endpoint_url(
            "/uclanlp/visualbert-nlvr2-coco-pre/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.visualbert.modeling
import unitorch.cli.models.visualbert.processing
from unitorch.cli.models.visualbert.modeling import (
    VisualBertForClassification,
    VisualBertForPretrain,
)
from unitorch.cli.models.visualbert.processing import VisualBertProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch import autocast

from unitorch.models.diffusers import VAEForDiffusion as _VAEForDiffusion
from unitorch.utils import (
    pop_value,
    nested_dict_value,
    is_bfloat16_available,
    is_cuda_available,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import DiffusionOutputs, LossOutputs
from unitorch.cli.models import diffusion_model_decorator
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
    load_weight,
)


@register_model("core/model/diffusers/vae")
class VAEForDiffusion(_VAEForDiffusion):
    def __init__(
        self,
        config_path: str,
        patch_size: Optional[int] = 32,
        stride: Optional[int] = 16,
        use_fp16: Optional[bool] = True,
        use_bf16: Optional[bool] = False,
    ):
        super().__init__(config_path, patch_size, stride)
        self.use_dtype = torch.float16 if use_fp16 else torch.float32
        self.use_dtype = (
            torch.bfloat16 if use_bf16 and is_bfloat16_available() else self.use_dtype
        )

    @classmethod
    @add_default_section_for_init("core/model/diffusers/vae")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/diffusers/vae")
        pretrained_name = config.getoption("pretrained_name", "stable-v1.5")
        pretrained_infos = nested_dict_value(pretrained_stable_infos, pretrained_name)

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_infos, "vae", "config"),
        )
        config_path = cached_path(config_path)

        patch_size = config.getoption("patch_size", 32)
        stride = config.getoption("stride", 16)
        use_fp16 = config.getoption("use_fp16", True)
        use_bf16 = config.getoption("use_bf16", False)

        inst = cls(
            config_path,
            patch_size=patch_size,
            stride=stride,
            use_fp16=use_fp16,
            use_bf16=use_bf16,
        )

        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_infos, "vae", "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    def forward(self, pixel_values):
        with autocast(
            device_type=("cuda" if torch.cuda.is_available() else "cpu"),
            dtype=self.use_dtype,
        ):
            loss = super().forward(pixel_values)
            return LossOutputs(loss=loss)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.controlnet_xl.text2image import (
    ControlNetXLForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.controlnet_xl.image2image import (
    ControlNetXLForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.controlnet_xl.inpainting import (
    ControlNetXLForImageInpaintingFastAPIPipeline,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import numpy as np
from PIL import Image, ImageFilter
from typing import Any, Dict, List, Optional, Tuple, Union
from unitorch.utils import is_opencv_available

from unitorch.cli.pipelines.stable_3.text2image import (
    Stable3ForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_3.image2image import (
    Stable3ForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_3.inpainting import (
    Stable3ForImageInpaintingPipeline,
)
